!******************************************************************************
!
! NS3D: a Navier-Stokes solver for incompressible and stratified flow
!
! Organization : LadHyX, Ecole Polytechnique, France
!
! Permission is granted to copy and distribute this file or modified
! versions of this file for no fee, provided the copyright notice and
! this permission notice are preserved on all copies.
!
!******************************************************************************


!***************************************************************************
!  FFT Module
!***************************************************************************
!****h* ns3d/fft
!  NAME
!    fft
!
!  DESCRIPTION
!    Implement common drivers for different FFT libraries: JMFFT-8.0, FFTW 3.01, ASL, ESSL
!    Every FFT library must be Initialized with init_fft() subroutine before performing
!    forward FFT with fwd_fft() or backward FFT with bck_fft()
!***



#if defined(JMFFT)
! ############################################################
!
!    JMFFT-8.0: a Fortran portable library by Jean-Marie Teuler, CNRS, teuler@lcp.u-psud.fr
!               for 3D/2D FFT real -> complex
! ############################################################

module fft

    implicit none

    private

    character(len=70), save, public :: fft_lib_name
    public :: init_fft, fwd_fft, bck_fft

    integer, save :: n1,n2,n3,d1,d2,d3
    double precision, save :: scaling

    integer, save :: lbnd1, ubnd1, lbnd2, ubnd2

    contains

! We include all the JMFFT library files
#include "JMFFT-8.0/src/jmccfft.f90"
#include "JMFFT-8.0/src/jmccfft2d.f90"
#include "JMFFT-8.0/src/jmccfft3d.f90"
#include "JMFFT-8.0/src/jmccfftm.f90"
#include "JMFFT-8.0/src/jmccm1d.f90"
#include "JMFFT-8.0/src/jmccm1d2.f90"
#include "JMFFT-8.0/src/jmccm1d3.f90"
#include "JMFFT-8.0/src/jmccm1d4.f90"
#include "JMFFT-8.0/src/jmccm1d5.f90"
#include "JMFFT-8.0/src/jmccm1dp.f90"
#include "JMFFT-8.0/src/jmcctranspcs.f90"
#include "JMFFT-8.0/src/jmcfftmlt.f90"
#include "JMFFT-8.0/src/jmcftfax.f90"
#include "JMFFT-8.0/src/jmcsfft.f90"
#include "JMFFT-8.0/src/jmcsfft2d.f90"
#include "JMFFT-8.0/src/jmcsfft3d.f90"
#include "JMFFT-8.0/src/jmcsfftm.f90"
#include "JMFFT-8.0/src/jmcsm1d.f90"
#include "JMFFT-8.0/src/jmcsm1dxy.f90"
#include "JMFFT-8.0/src/jmdecoup.f90"
#include "JMFFT-8.0/src/jmdecoup3.f90"
#include "JMFFT-8.0/src/jmerreur1.f90"
#include "JMFFT-8.0/src/jmerreur2.f90"
#include "JMFFT-8.0/src/jmerreur3.f90"
#include "JMFFT-8.0/src/jmerreur4.f90"
#include "JMFFT-8.0/src/jmfact.f90"
#include "JMFFT-8.0/src/jmfftfax.f90"
#include "JMFFT-8.0/src/jmgetcode.f90"
#include "JMFFT-8.0/src/jmgeterreur.f90"
#include "JMFFT-8.0/src/jmgeterror.f90"
#include "JMFFT-8.0/src/jmgetmessage.f90"
#include "JMFFT-8.0/src/jmgetnwork.f90"
#include "JMFFT-8.0/src/jmgetsetcode.f90"
#include "JMFFT-8.0/src/jmgetseterreur.f90"
#include "JMFFT-8.0/src/jmgetseterror.f90"
#include "JMFFT-8.0/src/jmgetsetnwork.f90"
#include "JMFFT-8.0/src/jmrfftmlt.f90"
#include "JMFFT-8.0/src/jmscfft.f90"
#include "JMFFT-8.0/src/jmscfft2d.f90"
#include "JMFFT-8.0/src/jmscfft3d.f90"
#include "JMFFT-8.0/src/jmscfftm.f90"
#include "JMFFT-8.0/src/jmscm1d.f90"
#include "JMFFT-8.0/src/jmscm1dxy.f90"
#include "JMFFT-8.0/src/jmsetcode.f90"
#include "JMFFT-8.0/src/jmseterreur.f90"
#include "JMFFT-8.0/src/jmseterror.f90"
#include "JMFFT-8.0/src/jmsetnwork.f90"
#include "JMFFT-8.0/src/jmtable.f90"
#include "JMFFT-8.0/src/jmtransp.f90"
#include "JMFFT-8.0/src/jmsinfftmlt.f90"
#include "JMFFT-8.0/src/jmcosfftmlt.f90"
#include "JMFFT-8.0/src/jmsinfftfax.f90"
#include "JMFFT-8.0/src/jmcosfftfax.f90"
#include "JMFFT-8.0/src/jmtablesin.f90"
#include "JMFFT-8.0/src/jmtablesincos.f90"

! *******************************************
    subroutine init_fft(b,an1,an2,an3,ad1,ad2,ad3,work)
! *******************************************
    !****f* fft/init_fft
    !  NAME
    !    init_fft
    !
    !  DESCRIPTION
    !    Initialize the FFT library
    !
    !  INPUTS
    !    b: field whose dimensions are those of the FFT that are to be performed later.
    !***

        implicit none

        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(*), intent(inout) :: work
        integer, intent(in) :: an1,an2,an3,ad1,ad2,ad3

        n1 = an1
        n2 = an2
        n3 = an3
        d1 = ad1
        d2 = ad2
        d3 = ad3

        ! n3 \= 1; we use a 3D FFT
        if (n3 /= 1) then
             fft_lib_name = 'JMFFT-8.0 3D (Author: Jean-Marie Teuler, CNRS)'

            scaling = 1./(n1*n2*n3)

            ! the library JMFFT requires two working arrays while only one is given in argument of the subroutine init_fft.
            ! We create artificially two arrays into our single array "work" by determining lower and upper bounds of inner arrays.
            lbnd1 = 1
            ubnd1= 100 + 2*(n1+n2+n3)
            lbnd2 = ubnd1+1
            ubnd2 = ubnd1 + 512*max(n1,n2,n3)

            call SCFFT3D(0,n1,n2,n3,scaling,b,d1,d2,b,d1/2,d2,work(lbnd1:ubnd1),work(lbnd2:ubnd2),0)

        ! n3=1; we use a 2D FFT
        else
            fft_lib_name = 'JMFFT-8.0 2D (Author: Jean-Marie Teuler, CNRS)'

            scaling = 1./(n1*n2)

            lbnd1 = 1
            ubnd1= 100 + 2*(n1+n2)
            lbnd2 = ubnd1+1
            ubnd2 = ubnd1 + 512*max(n1,n2)

            call SCFFT2D(0,n1,n2,scaling,b,d1,b,d1/2,work(lbnd1:ubnd1),work(lbnd2:ubnd2),0)
        end if


    end subroutine


! *******************************************
    subroutine fwd_fft(b,work)
! *******************************************
    !****f* fft/fwd_fft
    !  NAME
    !    fwd_fft
    !
    !  DESCRIPTION
    !    Perform a forward FFT
    !
    !  INPUTS
    !    b: field whose forward FFT is wanted
    !
    !  OUTPUT
    !    b: forward FFT of the input field
    !***

        implicit none

        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(*), intent(inout) :: work

        if (n3 /= 1) then
            call SCFFT3D(-1,n1,n2,n3,scaling,b,d1,d2,b,d1/2,d2,work(lbnd1:ubnd1),work(lbnd2:ubnd2),0)
        else
            call SCFFT2D(-1,n1,n2,scaling,b,d1,b,d1/2,work(lbnd1:ubnd1),work(lbnd2:ubnd2),0)
        end if

    end subroutine


! *******************************************
    subroutine bck_fft(b,work)
! *******************************************
    !****f* fft/bck_fft
    !  NAME
    !    bck_fft
    !
    !  DESCRIPTION
    !    Perform a backward FFT
    !
    !  INPUTS
    !    b: field whose backward FFT is wanted
    !
    !  OUTPUT
    !    b: backward FFT of the input field
    !***

        implicit none

        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(*), intent(inout) :: work

        if (n3 /= 1) then
            call CSFFT3D(1,n1,n2,n3,1.,b,d1/2,d2,b,d1,d2,work(lbnd1:ubnd1),work(lbnd2:ubnd2),0)
        else
            call CSFFT2D(1,n1,n2,1.,b,d1/2,b,d1,work(lbnd1:ubnd1),work(lbnd2:ubnd2),0)
        end if

    end subroutine

end module



#elif defined(FFTW) || defined(FFTW_OPENMP)
! #######################################################
!
!     Library FFTW - mono-threaded version for 3D (and 2D) FFT real -> complex
!
! #######################################################

module fft

    implicit none

    private

    character(len=70), save, public :: fft_lib_name
    public :: init_fft, fwd_fft, bck_fft

    ! plans and parameters used by FFTW
    integer(kind=8), save :: plan, plan_i
    integer, parameter :: FFTW_FORWARD=-1
    integer, parameter :: FFTW_BACKWARD=+1
    integer, parameter :: FFTW_MEASURE=0
    integer, parameter :: FFTW_EXHAUSTIVE=8
    integer, parameter :: FFTW_PATIENT=32
    integer, parameter :: FFTW_ESTIMATE=64

    integer, save :: n1,n2,n3,d1,d2,d3
    double precision, save :: scaling

    contains

! *******************************************
    subroutine init_fft(b,an1,an2,an3,ad1,ad2,ad3,work)
! *******************************************
    !****f* fft/init_fft
    !  NAME
    !    init_fft
    !
    !  DESCRIPTION
    !    Initialize the FFT library
    !
    !  INPUTS
    !    b: field whose dimensions are those of the FFT that are to be performed later.
    !    n1,n2,n3: logical dimensions of the FFT along the first, second and third directions.
    !              The FFT real -> complex which is the 1st of the 3 required 1D FFT is performed along the 1st direction.
    !              Because of conjugate relations, only half of the modes along this 1st direction are stored in spectral space.
    !    d1,d2,d3: physical dimensions of the array. d1 >= 2*(n1/2+1), d2 >= n2, d3 >= n3
    !    work: this argument is optional and not used here. It is kept only for compatibilty with other FFT interfaces.
    !***

        implicit none

        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(*), intent(in), optional :: work
        integer, intent(in) :: an1,an2,an3,ad1,ad2,ad3
        integer, dimension(0:2) :: ta, nembed, nembed_out
        logical :: trash

#ifdef FFTW_OPENMP
        integer return_thread

        call dfftw_init_threads(return_thread)
        call dfftw_plan_with_nthreads(FFTW_OPENMP)

        write(fft_lib_name,'(A38,I3)') 'FFTW 3.x OpenMP | Nb threads: ', FFTW_OPENMP
#else
        fft_lib_name = 'FFTW 3.x'
#endif

        ! Only to avoid the warning "The variable 'work' was not used" that is issued by some compilers.
        trash = present(work)

        n1 = an1
        n2 = an2
        n3 = an3
        d1 = ad1
        d2 = ad2
        d3 = ad3
        scaling = 1./(n1*n2*n3)

        ta(0) = n1
        ta(1) = n2
        ta(2) = n3
        nembed(0) = d1
        nembed(1) = d2
        nembed(2) = d3
        nembed_out(0) = d1/2
        nembed_out(1) = d2
        nembed_out(2) = d3

        call dfftw_plan_many_dft_r2c(plan,3,ta,1,b,nembed,1,1,b,nembed_out,1,1,FFTW_PATIENT)
        call dfftw_plan_many_dft_c2r(plan_i,3,ta,1,b,nembed_out,1,1,b,nembed,1,1,FFTW_PATIENT)

    end subroutine


! *******************************************
    subroutine fwd_fft(b,work)
! *******************************************
    !****f* fft/fwd_fft
    !  NAME
    !    fwd_fft
    !
    !  DESCRIPTION
    !    Perform a forward FFT
    !
    !  INPUTS
    !    b(x1,x2,x3): field in physical space
    !    work: this argument is optional and not used here. It is kept only for compatibilty with other FFT interfaces.
    !
    !  OUTPUT
    !    b(2,xk1,xk2,xk3): field in spectral space. Real and imaginary parts are stores contiguously.
    !***

        implicit none
        double precision, dimension(0:d1-1,0:d2-1,0:d3-1), intent(inout) :: b
        double precision, dimension(*), intent(in), optional :: work
        logical :: trash

        ! Only to avoid the warning "The variable 'work' was not used" that is issued by come compilers.
        trash = present(work)

        call dfftw_execute_dft_r2c(plan, b, b)

        ! we normalize the Fourier coefficients
        b = scaling*b

    end subroutine


! *******************************************
    subroutine bck_fft(b,work)
! *******************************************
    !****f* fft/bck_fft
    !  NAME
    !    bck_fft
    !
    !  DESCRIPTION
    !    Perform a backward FFT
    !
    !  INPUTS
    !    b(2,xk1,xk2,xk3): field in spectral space. Real and imaginary parts are stored contiguously.
    !    work: this argument is optional and not used here. It is kept only for compatibilty with other FFT interfaces.
    !
    !  OUTPUT
    !    b(x1,x2,x3): field in physical space
    !***

        implicit none

        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(*), intent(in), optional :: work
        logical :: trash

        ! Only to avoid the warning "The variable 'work' was not used" that is issued by some compilers.
        trash = present(work)

        call dfftw_execute_dft_c2r(plan_i,b,b)

    end subroutine

end module



#elif defined(MATHKEISAN)
! ############################################################
!
!    MathKeisan: fft library for NEC-SX8 for 3D/2D FFT real -> complex
!
! ############################################################

module fft

    implicit none

    private

    character(len=70), save, public :: fft_lib_name
    public :: init_fft, fwd_fft, bck_fft

    integer, save :: n1,n2,n3,d1,d2,d3
    double precision, save :: scaling

    integer, save :: lbnd1, ubnd1, lbnd2, ubnd2, lbnd3, ubnd3

    contains

! *******************************************
    subroutine init_fft(b,an1,an2,an3,ad1,ad2,ad3,work)
! *******************************************
    !****f* fft/init_fft
    !  NAME
    !    init_fft
    !
    !  DESCRIPTION
    !    Initialize the FFT library
    !
    !  INPUTS
    !    b: field whose dimensions are those of the FFT that are to be performed later.
    !***

        implicit none

        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(*), intent(inout) :: work
        integer, intent(in) :: an1,an2,an3,ad1,ad2,ad3

        n1 = an1
        n2 = an2
        n3 = an3
        d1 = ad1
        d2 = ad2
        d3 = ad3

        ! n3 \= 1; we use a 3D FFT
        if (n3 /= 1) then
            fft_lib_name = 'MathKeisan 3D (to check...)'

            scaling = 1./(n1*n2*n3)

            ! the library MathKeisan requires two working arrays while only one is given in argument of the subroutine init_fft.
            ! We create artificially different arrays into our single array "work" by determining lower
            ! and upper bounds of inner arrays.
            lbnd1 = 1
            ubnd1 = 2*(n1+n2+n3)+195
            lbnd2 = ubnd1 + 1
            ubnd2 = ubnd1 + 2*(n1+n2+n3)+195
            lbnd3 = ubnd2 + 1
            ubnd3 = ubnd2 + 4*256*max(n1,n2,n3)

            call DZFFT3D(0,n1,n2,n3,scaling,b,d1,d2,b,d1/2,d2,work(lbnd1:ubnd1),work(lbnd3:ubnd3),0)
            call ZDFFT3D(0,n1,n2,n3,scaling,b,d1/2,d2,b,d1,d2,work(lbnd2:ubnd2),work(lbnd3:ubnd3),0)


        ! n3=1; we use a 2D FFT
        else
            fft_lib_name = 'MathKeisan 2D'

            scaling = 1./(n1*n2)

            lbnd1 = 1
            ubnd1 = 2*(n1+n2)+130
            lbnd1 = ubnd1 + 1
            ubnd1 = ubnd1 + 2*(n1+n2)+130
            lbnd3 = ubnd2 + 1
            ubnd3 = ubnd2 + 6*n1*n2

            call DZFFT2D(0,n1,n2,scaling,b,d1,b,d1/2,work(lbnd1:ubnd1),work(lbnd3:ubnd3),0)
            call ZDFFT2D(0,n1,n2,scaling,b,d1/2,b,d1,work(lbnd2:ubnd2),work(lbnd3:ubnd3),0)
        end if


    end subroutine


! *******************************************
    subroutine fwd_fft(b,work)
! *******************************************
    !****f* fft/fwd_fft
    !  NAME
    !    fwd_fft
    !
    !  DESCRIPTION
    !    Perform a forward FFT
    !
    !  INPUTS
    !    b: field whose forward FFT is wanted
    !
    !  OUTPUT
    !    b: forward FFT of the input field
    !
    !  REM
    !    The sign in the exponential term for the forward and backward FFT
    !    is the inverse than usual
    !***

        implicit none

        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(*), intent(inout) :: work

        if (n3 /= 1) then
            call DZFFT3D(-1,n1,n2,n3,scaling,b,d1,d2,b,d1/2,d2,work(lbnd1:ubnd1),work(lbnd3:ubnd3),0)
        else
            call DZFFT2D(-1,n1,n2,scaling,b,d1,b,d1/2,work(lbnd1:ubnd1),work(lbnd3:ubnd3),0)
        end if

    end subroutine


! *******************************************
    subroutine bck_fft(b,work)
! *******************************************
    !****f* fft/bck_fft
    !  NAME
    !    bck_fft
    !
    !  DESCRIPTION
    !    Perform a backward FFT
    !
    !  INPUTS
    !    b: field whose backward FFT is wanted
    !
    !  OUTPUT
    !    b: backward FFT of the input field
    !
    !  REM
    !    The sign in the exponential term for the forward and backward FFT
    !    is the inverse than usual
    !***

        implicit none

        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(*), intent(inout) :: work

        if (n3 /= 1) then
            call ZDFFT3D(1,n1,n2,n3,1.,b,d1/2,d2,b,d1,d2,work(lbnd2:ubnd2),work(lbnd3:ubnd3),0)
        else
            call ZDFFT2D(1,n1,n2,1.,b,d1/2,b,d1,work(lbnd2:ubnd2),work(lbnd3:ubnd3),0)
        end if

    end subroutine

end module


#elif defined(ASL) || defined(ASL_PARALLEL)
! #######################################################
!
!       Library ASL
!
! #######################################################

module fft

    implicit none

    private

    character(len=70), save, public :: fft_lib_name
    public :: init_fft, fwd_fft, bck_fft

    integer, save :: n1,n2,n3,d1,d2,d3
    double precision, save :: scaling

    ! necessary for working storage
    integer, save :: lbnd1, ubnd1, lbnd2, ubnd2
    integer, dimension(60), save :: ifax

    contains

! *************************************************************************
    subroutine init_fft(b,an1,an2,an3,ad1,ad2,ad3,work)
! *************************************************************************
    !****f* fft/init_fft
    !  NAME
    !    init_fft
    !
    !  DESCRIPTION
    !    Initialize the FFT library
    !
    !  INPUTS
    !    b: field whose dimensions are those of the FFT that are to be performed later.
    !***

        implicit none

        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(*), intent(inout) :: work
        integer, intent(in) :: an1,an2,an3,ad1,ad2,ad3
        integer :: err

! we reserve the number of required threads for the following of the program
!PDIR RESERVE=ASL_PARALLEL

        n1 = an1
        n2 = an2
        n3 = an3
        d1 = ad1
        d2 = ad2
        d3 = ad3

        if (n3 /= 1) then
            fft_lib_name = 'ASL 3D'

            scaling = 1./(n1*n2*n3)

            ! we determine the lower and upper bounds of the two "internal" working arrays that are necessary
            lbnd1=1
            ubnd1=n1+2*(n2+n3)
            lbnd2=ubnd1+1
            ubnd2=ubnd1+d1*d2*d3

            ! we must set the working field to 0 otherwise the FFT sometimes bugs for some mysterious reasons...
            work(lbnd2:ubnd2) = 0.

            ! remark: the following subroutine does not work to set the "non-physical elements" to 0
            ! there is something I don't understand...
            !     call DFR3CL(b,n1,n2,n3,d1,d2,d3,err)

#ifdef ASL_PARALLEL
            call QFR3FB(n1,n2,n3,b,d1,d2,d3,0,ifax,work(lbnd1:ubnd1),work(lbnd2:ubnd2),ASL_PARALLEL,err)
#else
            call DFR3FB(n1,n2,n3,b,d1,d2,d3,0,ifax,work(lbnd1:ubnd1),work(lbnd2:ubnd2),err)
#endif

        else
            fft_lib_name = 'ASL 2D'

            scaling = 1./(n1*n2)

            lbnd1=1
            ubnd1=n1+2*n2
            lbnd2=ubnd1+1
            ubnd2=ubnd1+d1*d2

            ! we must set the working field to 0 otherwise the FFT sometimes bugs for some mysterious reasons...
            work(lbnd2:ubnd2) = 0.

#ifdef ASL_PARALLEL
            call QFR2FB(n1,n2,b,d1,d2,0,ifax,work(lbnd1:ubnd1),work(lbnd2:ubnd2),ASL_PARALLEL,err)
#else
            call DFR2FB(n1,n2,b,d1,d2,0,ifax,work(lbnd1:ubnd1),work(lbnd2:ubnd2),err)
#endif

        end if

    end subroutine


! *************************************************************************
    subroutine fwd_fft(b,work)
! *************************************************************************
    !****f* fft/fwd_fft
    !  NAME
    !    fwd_fft
    !
    !  DESCRIPTION
    !    Perform a forward FFT
    !
    !  INPUTS
    !    b: field whose forward FFT is wanted
    !
    !  OUTPUT
    !    b: forward FFT of the input field
    !***

        implicit none

        double precision, dimension(0:d1-1,0:d2-1,0:d3-1), intent(inout) :: b
        double precision, dimension(*), intent(inout) :: work
        integer :: err

        ! we must set the working field to 0 otherwise the FFT sometimes bugs for some mysterious reasons...
        work(lbnd2:ubnd2) = 0.

        ! we set to 0 the elements with no physical signification to avoid float number exception (cd ASL doc...)
        b(n1:d1-1,:,:) = 0.
        b(:,n2:d2-1,:) = 0.
        b(:,:,n3:d3-1) = 0.

#ifdef ASL_PARALLEL
        if (n3 /= 1) then
            call QFR3BF(n1,n2,n3,b,d1,d2,d3,1,ifax,work(lbnd1:ubnd1),work(lbnd2:ubnd2),ASL_PARALLEL,err)
        else
            call QFR2BF(n1,n2,b,d1,d2,1,ifax,work(lbnd1:ubnd1),work(lbnd2:ubnd2),ASL_PARALLEL,err)
        end if
#else
        if (n3 /= 1) then
            call DFR3BF(n1,n2,n3,b,d1,d2,d3,1,ifax,work(lbnd1:ubnd1),work(lbnd2:ubnd2),err)
        else
            call DFR2BF(n1,n2,b,d1,d2,1,ifax,work(lbnd1:ubnd1),work(lbnd2:ubnd2),err)
        end if
#endif

        ! we normalize the Fourier coefficients
        b = scaling*b

    end subroutine


! *************************************************************************
    subroutine bck_fft(b,work)
! *************************************************************************
    !****f* fft/bck_fft
    !  NAME
    !    bck_fft
    !
    !  DESCRIPTION
    !    Perform a backward FFT
    !
    !  INPUTS
    !    b: field whose backward FFT is wanted
    !
    !  OUTPUT
    !    b: backward FFT of the input field
    !***

        implicit none

        double precision, dimension(2,0:d1/2-1,0:d2-1,0:d3-1), intent(inout) :: b
        double precision, dimension(*), intent(inout) :: work
        integer :: err

        ! we must set the working field to 0 otherwise the FFT sometimes bugs for some mysterious reasons...
        work(lbnd2:ubnd2) = 0.

        ! we set to 0 the elements with no physical signification to avoid float number exception (cd ASL doc...)
        b(:,n1/2+1:d1/2-1,:,:) = 0.
        b(:,:,n2:d2-1,:) = 0.
        b(:,:,:,n3:d3-1) = 0.

#ifdef ASL_PARALLEL
        if (n3 /= 1) then
            call QFR3BF(n1,n2,n3,b,d1,d2,d3,-1,ifax,work(lbnd1:ubnd1),work(lbnd2:ubnd2),ASL_PARALLEL,err)
        else
            call QFR2BF(n1,n2,b,d1,d2,-1,ifax,work(lbnd1:ubnd1),work(lbnd2:ubnd2),ASL_PARALLEL,err)
        end if
#else
        if (n3 /= 1) then
            call DFR3BF(n1,n2,n3,b,d1,d2,d3,-1,ifax,work(lbnd1:ubnd1),work(lbnd2:ubnd2),err)
        else
            call DFR2BF(n1,n2,b,d1,d2,-1,ifax,work(lbnd1:ubnd1),work(lbnd2:ubnd2),err)
        end if
#endif

    end subroutine

end module



#elif defined(ESSL)
! #######################################################
!
!       Library IBM ESSL - FFT real -> complex
!
! #######################################################

module fft

    implicit none

    private

    character(len=70), save, public :: fft_lib_name
    public :: init_fft, fwd_fft, bck_fft

    integer, save :: n1,n2,n3,d1,d2,d3
    double precision, save :: scaling

    ! useful for the 2D FFT to build working arrays
    integer, save :: naux1
    integer, save :: lbnd1, ubnd1, lbnd2, ubnd2

    contains


! *************************************************************************
    subroutine init_fft(b,an1,an2,an3,ad1,ad2,ad3,work)
! *************************************************************************
    !****f* fft/init_fft
    !  NAME
    !    init_fft
    !
    !  DESCRIPTION
    !    Initialize the FFT library
    !
    !  INPUTS
    !    b: field whose dimensions are those of the FFT that are to be performed later.
    !***

        implicit none

        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(*), intent(out), optional :: work
        integer, intent(in) :: an1,an2,an3,ad1,ad2,ad3
        double precision :: aux

        n1 = an1
        n2 = an2
        n3 = an3
        d1 = ad1
        d2 = ad2
        d3 = ad3

        if (n3 /= 1) then
            fft_lib_name = 'ESSL 3D'

            scaling = 1./(n1*n2*n3)
        else
            fft_lib_name = 'ESSL 2D'

            scaling = 1./(n1*n2)

            ! we determine the required size for the working array
            ! The ESSL documentation indicates that for a 2D real to complex FFT, the following storage is required:
            ! If n <= 2048, use naux1 = 42000.
            ! If n > 2048, use naux1 = 40000+1.64n1+2.28n2,
            ! where n = max(n1/2, n2).
            if (max(n1/2,n2) <= 2048) then
                naux1 = 42000
            else
                naux1 = 40000+1.64*n1+2.28*n2
            end if

            lbnd1 = 1
            ubnd1 = naux1
            lbnd2 = naux1+1
            ubnd2 = 2*naux1

            call DRCFT2(1,b,d1,b,d1/2,n1,n2,1,scaling,work(lbnd1:ubnd1),naux1,aux,0)
            call DCRFT2(1,b,d1/2,b,d1,n1,n2,-1,1.0,work(lbnd2:ubnd2),naux1,aux,0)
        endif

    end subroutine


! *************************************************************************
    subroutine fwd_fft(b,work)
! *************************************************************************
    !****f* fft/fwd_fft
    !  NAME
    !    fwd_fft
    !
    !  DESCRIPTION
    !    Perform a forward FFT
    !
    !  INPUTS
    !    b: field whose forward FFT is wanted
    !
    !  OUTPUT
    !    b: forward FFT of the input field
    !***

        implicit none

        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(*), intent(inout) :: work
        double precision :: aux

        if (n3 /= 1) then
            call DRCFT3(b,d1,d1*d2,b,d1/2,d1/2*d2,n1,n2,n3,1,scaling,aux,0)
        else
            call DRCFT2(0,b,d1,b,d1/2,n1,n2,1,scaling,work(lbnd1:ubnd1),naux1,aux,0)
        end if

    end subroutine


! *************************************************************************
    subroutine bck_fft(b, work)
! *************************************************************************
    !****f* fft/bck_fft
    !  NAME
    !    bck_fft
    !
    !  DESCRIPTION
    !    Perform a backward FFT
    !
    !  INPUTS
    !    b: field whose backward FFT is wanted
    !
    !  OUTPUT
    !    b: backward FFT of the input field
    !***

        implicit none
        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(*), intent(inout) :: work
        double precision :: aux

        if (n3 /= 1) then
            call DCRFT3(b,d1/2,d1/2*d2,b,d1,d1*d2,n1,n2,n3,-1,1.0,aux,0)
        else
            call DCRFT2(0,b,d1/2,b,d1,n1,n2,-1,1.0,work(lbnd2:ubnd2),naux1,aux,0)
        end if

    end subroutine

end module



#elif defined(FFTW_MPI) || defined(DFFTW_MPI_OPENMP)
! #######################################################
!
!     Library FFTW MPI - MPI version for 3D (and 2D) FFT real -> complex
!                        Warning: this is a "transposed"  3D FFT ; the 3 dimensions are transposed between the physical and
!                        spectral space. More details are given in the headers of bck_fft and fwd_fft subroutines.
! #######################################################

module fft

    implicit none

    private

    character(len=70), save, public :: fft_lib_name
    public :: init_fft, fwd_fft, bck_fft, init_dealiasing

    ! plans and parameters used by FFTW
    integer(kind=8), save :: plan_fwd1, plan_fwd2, plan_fwd3, plan_bck1, plan_bck2, plan_bck3
    integer, parameter :: FFTW_FORWARD=-1
    integer, parameter :: FFTW_BACKWARD=+1
    integer, parameter :: FFTW_MEASURE=0
    integer, parameter :: FFTW_EXHAUSTIVE=8
    integer, parameter :: FFTW_PATIENT=32
    integer, parameter :: FFTW_ESTIMATE=64

    ! Logical dimensions of the 3D FFT
    integer, save :: n1,n2,n3

    ! Physical dimensions of the array
    integer, save :: d1,d2,d3

    ! number of MPI processes
    integer, save :: np

    ! scaling factor
    double precision, save :: scaling

    ! number of modes and physical dimensions in the spectral space
    integer, save :: nk1,nk2,nk3,dk1,dk2,dk3

    ! In MPI parallelism, the data are splitted between the processes. Each processor has locally only a fraction of the array,
    ! of the modes etc. All the quantities below correspond to these "local" numbers.
    integer, save :: pn1, pn2, pn3
    integer, save :: pnk1, pnk2, pnk3
    integer, save :: pd1, pd2, pd3
    integer, save :: pdk1, pdk2, pdk3

    ! when de-aliasing is used, all the FFT do not have to be calculated. These array store the modes to calculate.
    logical, save :: dealiasing
    logical, dimension(:,:), allocatable :: trunc_k1k2, trunc_k2k1
    logical, dimension(:), allocatable :: trunc_k1

    include "mpif.h"


    contains

! *******************************************
    subroutine init_fft(b,an1,an2,an3,ad1,ad2,ad3,anp,work)
! *******************************************
    !****f* fft/init_fft
    !  NAME
    !    init_fft
    !
    !  DESCRIPTION
    !    Initialize the FFT library
    !
    !  INPUTS
    !    b: field whose dimensions are those of the FFT that are to be performed later.
    !    an1,an2,an3: logical dimensions of the FFT along the first, second and third directions.
    !                 The FFT real -> complex which is the 1st of the 3 required 1D FFT is performed along the 1st direction.
    !                 Because of conjugate relations, only half of the modes along this 1st direction are stored in spectral space.
    !
    !    ad1,ad2,ad3: physical dimensions of the array. d1 >= 2*(n1/2+1), d2 >= n2, d3 >= n3
    !    work: working array whose size must be >= 2*d1*d2*d3
    !    np: number of MPI processes
    !***

        implicit none

        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(*), intent(out) :: work
        integer, intent(in) :: an1,an2,an3,ad1,ad2,ad3,anp
        double precision, dimension(2,0:ad1/2-1) :: tk1
        double precision, dimension(2,0:ad2-1) :: tk2,qk2
        double precision, dimension(2,0:ad3-1) :: tk3,qk3

        ! constant parameters
        integer, parameter :: FFTW_FORWARD=-1
        integer, parameter :: FFTW_BACKWARD=+1
        integer, parameter :: FFTW_MEASURE=0
        integer, parameter :: FFTW_EXHAUSTIVE=8
        integer, parameter :: FFTW_PATIENT=32
        integer, parameter :: FFTW_ESTIMATE=64

#ifdef DFFTW_MPI_OPENMP
        integer return_thread

        call dfftw_init_threads(return_thread)
        call dfftw_plan_with_nthreads(DFFTW_MPI_OPENMP)

        write(fft_lib_name,'(A38,I3)') 'FFTW 3.x OpenMP/MPI 3D FFT | Nb threads: ', DFFTW_MPI_OPENMP
#else
        fft_lib_name = 'FFTW 3.x - MPI 3D transposed FFT'
#endif

        n1=an1
        n2=an2
        n3=an3
        d1=ad1
        d2=ad2
        d3=ad3
        np=anp

        scaling = 1./(n1*n2*n3)

        dealiasing = .false.

        nk1=n1/2+1
        nk2=n2
        nk3=n3
        dk1=d1/2
        dk2=d2
        dk3=d3

        pn1=n1/np
        pn2=n2/np
        pn3=n3/np
        pnk1=nk1/np
        pnk2=nk2/np
        pnk3=nk3/np

        pd1=d1/np
        pd2=d2/np
        pd3=d3/np
        pdk1=dk1/np
        pdk2=dk2/np
        pdk3=dk3/np

        call dfftw_plan_dft_r2c_1d( plan_fwd1,  n1,    b,  tk1,                 FFTW_PATIENT)
        call dfftw_plan_dft_1d(     plan_fwd2,  n2,    b,  tk2, FFTW_FORWARD,   FFTW_PATIENT)
        call dfftw_plan_dft_1d(     plan_fwd3,  n3,    tk3, qk3, FFTW_FORWARD,   FFTW_PATIENT)

        call dfftw_plan_dft_1d(     plan_bck1,  n3, b,  tk3, FFTW_BACKWARD,  FFTW_PATIENT)
        call dfftw_plan_dft_1d(     plan_bck2,  n2,  tk2, qk2, FFTW_BACKWARD,  FFTW_PATIENT)
        call dfftw_plan_dft_c2r_1d( plan_bck3,  n1,  b,  b,                  FFTW_PATIENT)

    end subroutine

! *******************************************
    subroutine init_dealiasing(active, trunc_k3k1k2)
! *******************************************
    !****f* fft/init_dealiasing
    !  NAME
    !    init_dealiasing
    !
    !  DESCRIPTION
    !    Activate or deactivate the calculation of modes that are not useful because of dealiasing. It can save time
    !    if a large number of modes are truncated.
    !
    !  INPUTS
    !    active: .true. (dealiasing on) or .false. (dealiasing off)
    !    trunc_k3k1k2: array that contains the modes that are kept in spectral space (.true.) or truncated (.false.)
    !                  the order is the "transposed" order. So if the field is (x1,x2,x3/np) in physical space, the
    !                  array trunc_k3k1k2 is given in order (xk3,xk1,xk2/np)
    !***

        implicit none

        logical, intent(in) :: active
        logical, dimension(0:dk3-1,0:dk1-1,0:pdk2-1), intent(in), optional :: trunc_k3k1k2
        logical, dimension(0:dk1-1) :: tmp_trunc_k1
        integer :: code

        if (allocated(trunc_k1k2)) then
            deallocate(trunc_k1k2)
        end if

        if (allocated(trunc_k2k1)) then
            deallocate(trunc_k2k1)
        end if

        if (allocated(trunc_k1)) then
            deallocate(trunc_k1)
        end if


        if (active .and. present(trunc_k3k1k2)) then
            dealiasing = .true.

            allocate(trunc_k1k2(0:dk1-1,0:pdk2-1))
            allocate(trunc_k2k1(0:pdk2-1,0:dk1-1))
            allocate(trunc_k1(0:dk1-1))

            ! we determine which modes should be calculated in the plane (k1,pk2).
            ! If any mode (k1,pk2) along direction k3 is .true. then the corresponding k3-1D-FFT should be calculated.
            ! The transposition of trunc_k1k2 is stored only to allow a better cache memory access in bck_fft_step1
            trunc_k1k2 = any(trunc_k3k1k2,1)
            trunc_k2k1 = transpose(trunc_k1k2)

            ! we determine which modes should be calculated on the axis k1.
            ! If any mode k1 along direction k2 is .true. in the array trunc_k1k2 then the corresponding k2-1D FFT should be done.
            ! We start with a local reduction (logical .and.) followed by a global reduction between processes.
            tmp_trunc_k1 = any(trunc_k1k2,2)
            call Mpi_allreduce(tmp_trunc_k1,trunc_k1,dk1,MPI_LOGICAL,MPI_LOR ,MPI_COMM_WORLD ,code)

        else
            dealiasing = .false.
        end if

    end subroutine


! *******************************************
    subroutine fwd_fft(b,work)
! *******************************************
    !****f* fft/fwd_fft
    !  NAME
    !    fwd_fft
    !
    !  DESCRIPTION
    !    Perform a forward FFT
    !
    !  INPUTS
    !    b(x1,x2,x3/np): field in the physical space. The third dimension is splitted between the different processes.
    !    work: working array whose size must be >= 2*d1*d2*d3
    !
    !  OUTPUT
    !    b(2,xk3,xk1,xk2/np): field in the spectral space.
    !                         The real and imaginary part are stored contiguously.
    !                         This is a "transposed" FFT: dimensions 1,2 and 3 are not in the same order in the spectral and
    !                         physical space. The last dimension is splitted between the different processes.
    !***

        implicit none
        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(0:d1-1,0:d2-1,0:pd3-1,2), intent(out) :: work
        integer :: code

        call fwd_fft_step1(b,work(:,:,:,1))

        call fwd_fft_step2(work(:,:,:,1),work(:,:,:,2))

        call MPI_alltoall(work(:,:,:,2),d1*d2*d3/np**2,MPI_DOUBLE_PRECISION,work(:,:,:,1),d1*d2*d3/np**2, &
                          MPI_DOUBLE_PRECISION,MPI_COMM_WORLD,code)

        call fwd_fft_step3(work(:,:,:,1),b)

    end subroutine


! *******************************************
    subroutine bck_fft(b,work)
! *******************************************
    !****f* fft/bck_fft
    !  NAME
    !    bck_fft
    !
    !  DESCRIPTION
    !    Perform a backward FFT. This is the exact inverse of fwd_fft i.e. ffw_fft(bck_fft(a)) = a
    !
    !  INPUTS
    !    b(2,xk3,xk1,xk2/np): field in the spectral space.
    !                         The real and imaginary part are stored contiguously.
    !                         The third dimension is splitted between the different processes.
    !
    !    work: working array whose size must be >= 2*d1*d2*d3
    !
    !  OUTPUT
    !    b(x1,x2,x3/np): field in the physical space.
    !                    The third dimension is splitted between the different processes.
    !***

        implicit none

        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(0:d1-1,0:d2-1,0:pd3-1,2), intent(out) :: work
        integer :: code

        call bck_fft_step1(b,work(:,:,:,1))

        call MPI_alltoall(work(:,:,:,1),d1*d2*d3/np**2,MPI_DOUBLE_PRECISION,work(:,:,:,2),d1*d2*d3/np**2, &
                          MPI_DOUBLE_PRECISION,MPI_COMM_WORLD,code)

        call bck_fft_step2(work(:,:,:,2),b)

        call bck_fft_step3(b)

    end subroutine

! *******************************************
    subroutine fwd_fft_step1(a,c)
! *******************************************

        implicit none

        double precision, dimension(0:d1-1,0:d2-1,0:pd3-1), intent(in) :: a
        double precision, dimension(2,0:d2-1,0:dk1-1,0:pd3-1), intent(out) :: c
        double precision, dimension(2,0:dk1-1) :: b2
        integer :: ik1,i2,i3

        do i3=0,pn3-1
            do i2=0,n2-1
                call dfftw_execute_dft_r2c(plan_fwd1, a(0,i2,i3), b2)

                ! we transpose the subarray that has just been FFTed while the data are still in the cache
                do ik1=0,nk1-1
                    c(:,i2,ik1,i3) = b2(:,ik1)
                end do
            end do
        end do

    end subroutine


! *******************************************
    subroutine fwd_fft_step2(a,c)
! *******************************************

        implicit none

        double precision, dimension(2,0:d2-1,0:dk1-1,0:pd3-1), intent(in) :: a
        double precision, dimension(2,0:pd3-1,0:dk1-1,0:pdk2-1,0:np-1), intent(out) :: c
        double precision, dimension(2,0:dk2-1) :: b2
        integer :: i3,ik1,ik2,ip

        if (dealiasing) then
            do ik1=0,nk1-1
                if (trunc_k1(ik1)) then
                    do i3=0,pn3-1
                        call dfftw_execute_dft(plan_fwd2, a(1,0,ik1,i3), b2)

                        do ik2=0,pnk2-1
                            do ip=0,np-1
                                c(:,i3,ik1,ik2,ip) = b2(:,ip+np*ik2)
                            end do
                        end do
                    end do
                end if
            end do
        else
            do ik1=0,nk1-1
                do i3=0,pn3-1
                    call dfftw_execute_dft(plan_fwd2, a(1,0,ik1,i3), b2)

                    do ik2=0,pnk2-1
                        do ip=0,np-1
                            c(:,i3,ik1,ik2,ip) = b2(:,ip+np*ik2)
                        end do
                    end do
                end do
            end do
        end if

    end subroutine


! *******************************************
    subroutine fwd_fft_step3(a,c)
! *******************************************

        implicit none

        double precision, dimension(2,0:pd3-1,0:dk1-1,0:pdk2-1,0:np-1), intent(in) :: a
        double precision, dimension(2,0:dk3-1,0:dk1-1,0:pdk2-1), intent(out) :: c
        double precision, dimension(2,0:dk3-1) :: b2,b3
        integer :: ik1,ik2,ik3,i3,ip

        if (dealiasing) then
            do ik2=0,pnk2-1
                do ik1=0,nk1-1
                    if (trunc_k1k2(ik1,ik2)) then
                        do ip=0,np-1
                            do i3=0,pn3-1
                                b2(:,i3+ip*pn3) = a(:,i3,ik1,ik2,ip)
                            end do
                        end do

                        call dfftw_execute_dft(plan_fwd3, b2, b3)

                        do ik3=0,nk3-1
                            c(:,ik3,ik1,ik2) = scaling*b3(:,ik3)
                        end do
                    end if
                end do
            end do
        else
            do ik2=0,pnk2-1
                do ik1=0,nk1-1

                    do ip=0,np-1
                        do i3=0,pn3-1
                            b2(:,i3+ip*pn3) = a(:,i3,ik1,ik2,ip)
                        end do
                    end do

                    call dfftw_execute_dft(plan_fwd3, b2, b3)

                    do ik3=0,nk3-1
                        c(:,ik3,ik1,ik2) = scaling*b3(:,ik3)
                    end do
                end do
            end do
        end if
    end subroutine


! *******************************************
    subroutine bck_fft_step1(a,c)
! *******************************************

        implicit none

        double precision, dimension(2,0:dk3-1,0:dk1-1,0:pdk2-1), intent(in) :: a
        double precision, dimension(2,0:pdk2-1,0:dk1-1,0:pd3-1,0:np-1), intent(out) :: c
        double precision, dimension(2,0:dk3-1) :: b2
        integer :: i3,ik1,ik2,ip

        if (dealiasing) then

            ! It insures that the modes that are originally truncated in the array a remain truncated after the 1D-FFT
            c = 0.

            do ik1=0, nk1-1
                do ik2=0,pnk2-1
                    if (trunc_k2k1(ik2,ik1)) then
                        call dfftw_execute_dft(plan_bck1, a(1,0,ik1,ik2), b2)

                        ! we transpose the subarray that has just been FFTed while the data are still in the cache
                        do ip=0,np-1
                            do i3=0,pn3-1
                                c(:,ik2,ik1,i3,ip) = b2(:,i3+ip*pn3)
                            end do
                        end do
                    end if
                end do
            end do
        else
            do ik1=0, nk1-1
                do ik2=0,pnk2-1
                    call dfftw_execute_dft(plan_bck1, a(1,0,ik1,ik2), b2)

                    ! we transpose the subarray that has just been FFTed while the data are still in the cache
                    do ip=0,np-1
                        do i3=0,pn3-1
                            c(:,ik2,ik1,i3,ip) = b2(:,i3+ip*pn3)
                        end do
                    end do
                end do
            end do
        end if

    end subroutine


! *******************************************
    subroutine bck_fft_step2(a,c)
! *******************************************

        implicit none

        double precision, dimension(2,0:pdk2-1,0:dk1-1,0:pd3-1,0:np-1), intent(in) :: a
        double precision, dimension(2,0:dk1-1,0:d2-1,0:pd3-1), intent(out) :: c
        double precision, dimension(2,0:dk2-1) :: b2,b3
        integer :: i2,i3,ik1,ik2,ip

        if (dealiasing) then

            ! It insures that the modes that are originally truncated in the array a remain truncated after the 1D-FFT
            c = 0.

            do i3=0,pn3-1
                do ik1=0,nk1-1
                    if (trunc_k1(ik1)) then
                        do ip=0,np-1
                            do ik2=0,pnk2-1
                                b2(:,ip+np*ik2) = a(:,ik2,ik1,i3,ip)
                            end do
                        end do

                        call dfftw_execute_dft(plan_bck2,b2,b3)

                        do i2=0,n2-1
                            c(:,ik1,i2,i3) = b3(:,i2)
                        end do
                    end if
                end do
           end do
        else
            do i3=0,pn3-1
                do ik1=0,nk1-1

                    do ip=0,np-1
                        do ik2=0,pnk2-1
                            b2(:,ip+np*ik2) = a(:,ik2,ik1,i3,ip)
                        end do
                    end do

                    call dfftw_execute_dft(plan_bck2,b2,b3)

                    do i2=0,n2-1
                        c(:,ik1,i2,i3) = b3(:,i2)
                    end do
                end do
           end do
        end if

    end subroutine


! *******************************************
    subroutine bck_fft_step3(a)
! *******************************************

        implicit none

        double precision, dimension(2,0:dk1-1,0:d2-1,0:pd3-1), intent(inout) :: a
        integer :: i2,i3

        do i3=0,pn3-1
            do i2=0,n2-1
                call dfftw_execute_dft_c2r(plan_bck3, a(1,0,i2,i3), a(1,0,i2,i3))
            end do
        end do

    end subroutine

end module



#elif defined(MATHKEISAN_MPI)
! #######################################################
!
!     Library MATHKEISAN MPI - MPI version for 3D (and 2D) FFT using the MathKeisan library real -> complex
!                              Warning: this is a "transposed"  3D FFT ; the 3 dimensions are transposed between the physical and
!                              spectral space. More details are given in the headers of bck_fft and fwd_fft subroutines.
! #######################################################

module fft

    implicit none

    private

    character(len=70), save, public :: fft_lib_name
    public :: init_fft, fwd_fft, bck_fft, init_dealiasing

    ! Logical dimensions of the 3D FFT
    integer, save :: n1,n2,n3

    ! Physical dimensions of the array
    integer, save :: d1,d2,d3

    ! bounds of the working arrays
    integer, save :: lbnd0,ubnd0,lbnd1,ubnd1,lbnd2,ubnd2,lbnd3,ubnd3,lbnd4,ubnd4,lbnd5,ubnd5

    ! number of MPI processes
    integer, save :: np

    ! scaling factor
    double precision, save :: scaling

    ! number of modes and physical dimensions in the spectral space
    integer, save :: nk1,nk2,nk3,dk1,dk2,dk3

    ! In MPI parallelism, the data are splitted between the processes. Each processor has locally only a fraction of the array,
    ! of the modes etc. All the quantities below correspond to these "local" numbers.
    integer, save :: pn1, pn2, pn3
    integer, save :: pnk1, pnk2, pnk3
    integer, save :: pd1, pd2, pd3
    integer, save :: pdk1, pdk2, pdk3

    ! when de-aliasing is used, all the FFT do not have to be calculated. These array store the modes to calculate.
    logical, save :: dealiasing
    logical, dimension(:,:), allocatable :: trunc_k1k2, trunc_k2k1
    logical, dimension(:), allocatable :: trunc_k1

    include "mpif.h"


    contains

! *******************************************
    subroutine init_fft(b,an1,an2,an3,ad1,ad2,ad3,anp,work)
! *******************************************
    !****f* fft/init_fft
    !  NAME
    !    init_fft
    !
    !  DESCRIPTION
    !    Initialize the FFT library
    !
    !  INPUTS
    !    b: field whose dimensions are those of the FFT that are to be performed later.
    !    an1,an2,an3: logical dimensions of the FFT along the first, second and third directions.
    !                 The FFT real -> complex which is the 1st of the 3 required 1D FFT is performed along the 1st direction.
    !                 Because of conjugate relations, only half of the modes along this 1st direction are stored in spectral space.
    !
    !    ad1,ad2,ad3: physical dimensions of the array. d1 >= 2*(n1/2+1), d2 >= n2, d3 >= n3
    !    work: working array whose size must be >= 2*d1*d2*d3/np + 2*(2*n1+n2+n3)+4*64 + 4*max(n1,n2,n3)
    !    np: number of MPI processes
    !***

        implicit none

        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(*), intent(out) :: work
        integer, intent(in) :: an1,an2,an3,ad1,ad2,ad3,anp

        fft_lib_name = 'MathKeisan - MPI 3D transposed FFT'

        n1=an1
        n2=an2
        n3=an3
        d1=ad1
        d2=ad2
        d3=ad3
        np=anp

        scaling = 1./(n1*n2*n3)

        dealiasing = .false.

        nk1=n1/2+1
        nk2=n2
        nk3=n3
        dk1=d1/2
        dk2=d2
        dk3=d3

        pn1=n1/np
        pn2=n2/np
        pn3=n3/np
        pnk1=nk1/np
        pnk2=nk2/np
        pnk3=nk3/np

        pd1=d1/np
        pd2=d2/np
        pd3=d3/np
        pdk1=dk1/np
        pdk2=dk2/np
        pdk3=dk3/np

        ! the library MathKeisan requires several working arrays while only one is given in argument of the subroutine init_fft.
        ! We create artificially two arrays into our single array "work" by determining lower and upper bounds of inner arrays.
        lbnd0 = 1
        ubnd0 = 2*d1*d2*pd3

        lbnd1 = ubnd0 + 1
        ubnd1 = ubnd0 + 2*n1+64

        lbnd2 = ubnd1 + 1
        ubnd2 = ubnd1 + 2*n2+64

        lbnd3 = ubnd2 + 1
        ubnd3 = ubnd2 + 2*n3+64

        lbnd4 = ubnd3 + 1
        ubnd4 = ubnd3 + 2*n1+64

        lbnd5 = ubnd4 + 1
        ubnd5 = ubnd4 + 4*max(n1,n2,n3)

        call dzfft(0, n1, 1., b, b, work(lbnd1:ubnd1), work(lbnd5:ubnd5), 0)
        call zzfft(0, n2, 1., b, b, work(lbnd2:ubnd2), work(lbnd5:ubnd5), 0)
        call zzfft(0, n3, 1., b, b, work(lbnd3:ubnd3), work(lbnd5:ubnd5), 0)
        call zdfft(0, n1, 1., b, b, work(lbnd4:ubnd4), work(lbnd5:ubnd5), 0)

    end subroutine

! *******************************************
    subroutine init_dealiasing(active, trunc_k3k1k2)
! *******************************************
    !****f* fft/init_dealiasing
    !  NAME
    !    init_dealiasing
    !
    !  DESCRIPTION
    !    Activate or deactivate the calculation of modes that are not useful because of dealiasing. It can save time
    !    if a large number of modes are truncated.
    !
    !  INPUTS
    !    active: .true. (dealiasing on) or .false. (dealiasing off)
    !    trunc_k3k1k2: array that contains the modes that are kept in spectral space (.true.) or truncated (.false.)
    !                  the order is the "transposed" order. So if the field is (x1,x2,x3/np) in physical space, the
    !                  array trunc_k3k1k2 is given in order (xk3,xk1,xk2/np)
    !***

        implicit none

        logical, intent(in) :: active
        logical, dimension(0:dk3-1,0:dk1-1,0:pdk2-1), intent(in), optional :: trunc_k3k1k2
        logical, dimension(0:dk1-1) :: tmp_trunc_k1
        integer :: code

        if (allocated(trunc_k1k2)) then
            deallocate(trunc_k1k2)
        end if

        if (allocated(trunc_k2k1)) then
            deallocate(trunc_k2k1)
        end if

        if (allocated(trunc_k1)) then
            deallocate(trunc_k1)
        end if


        if (active .and. present(trunc_k3k1k2)) then
            dealiasing = .true.

            allocate(trunc_k1k2(0:dk1-1,0:pdk2-1))
            allocate(trunc_k2k1(0:pdk2-1,0:dk1-1))
            allocate(trunc_k1(0:dk1-1))

            ! we determine which modes should be calculated in the plane (k1,pk2).
            ! If any mode (k1,pk2) along direction k3 is .true. then the corresponding k3-1D-FFT should be calculated.
            ! The transposition of trunc_k1k2 is stored only to allow a better cache memory access in bck_fft_step1
            trunc_k1k2 = any(trunc_k3k1k2,1)
            trunc_k2k1 = transpose(trunc_k1k2)

            ! we determine which modes should be calculated on the axis k1.
            ! If any mode k1 along direction k2 is .true. in the array trunc_k1k2 then the corresponding k2-1D FFT should be done.
            ! We start with a local reduction (logical .and.) followed by a global reduction between processes.
            tmp_trunc_k1 = any(trunc_k1k2,2)
            call Mpi_allreduce(tmp_trunc_k1,trunc_k1,dk1,MPI_LOGICAL,MPI_LOR ,MPI_COMM_WORLD ,code)

        else
            dealiasing = .false.
        end if

    end subroutine


! *******************************************
    subroutine fwd_fft(b,work)
! *******************************************
    !****f* fft/fwd_fft
    !  NAME
    !    fwd_fft
    !
    !  DESCRIPTION
    !    Perform a forward FFT
    !
    !  INPUTS
    !    b(x1,x2,x3/np): field in the physical space. The third dimension is splitted between the different processes.
    !    work: working array
    !
    !  OUTPUT
    !    b(2,xk3,xk1,xk2/np): field in the spectral space.
    !                         The real and imaginary part are stored contiguously.
    !                         This is a "transposed" FFT: dimensions 1,2 and 3 are not in the same order in the spectral and
    !                         physical space. The last dimension is splitted between the different processes.
    !***

        implicit none
        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(0:d1-1,0:d2-1,0:pd3-1,*), intent(inout) :: work
        integer :: code

        call fwd_fft_step1(b,work(:,:,:,1),work)

        call fwd_fft_step2(work(:,:,:,1),work(:,:,:,2),work)

        call MPI_alltoall(work(:,:,:,2),d1*d2*d3/np**2,MPI_DOUBLE_PRECISION,work(:,:,:,1),d1*d2*d3/np**2, &
                          MPI_DOUBLE_PRECISION,MPI_COMM_WORLD,code)

        call fwd_fft_step3(work(:,:,:,1),b,work)

    end subroutine


! *******************************************
    subroutine bck_fft(b,work)
! *******************************************
    !****f* fft/bck_fft
    !  NAME
    !    bck_fft
    !
    !  DESCRIPTION
    !    Perform a backward FFT. This is the exact inverse of fwd_fft i.e. ffw_fft(bck_fft(a)) = a
    !
    !  INPUTS
    !    b(2,xk3,xk1,xk2/np): field in the spectral space.
    !                         The real and imaginary part are stored contiguously.
    !                         The third dimension is splitted between the different processes.
    !
    !    work: working array
    !
    !  OUTPUT
    !    b(x1,x2,x3/np): field in the physical space.
    !                    The third dimension is splitted between the different processes.
    !***

        implicit none

        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(0:d1-1,0:d2-1,0:pd3-1,*), intent(inout) :: work
        integer :: code

        call bck_fft_step1(b,work(:,:,:,1),work)

        call MPI_alltoall(work(:,:,:,1),d1*d2*d3/np**2,MPI_DOUBLE_PRECISION,work(:,:,:,2),d1*d2*d3/np**2, &
                          MPI_DOUBLE_PRECISION,MPI_COMM_WORLD,code)

        call bck_fft_step2(work(:,:,:,2),b,work)

        call bck_fft_step3(b,work)

    end subroutine

! *******************************************
    subroutine fwd_fft_step1(a,c,work)
! *******************************************

        implicit none

        double precision, dimension(0:d1-1,0:d2-1,0:pd3-1), intent(in) :: a
        double precision, dimension(2,0:d2-1,0:dk1-1,0:pd3-1), intent(out) :: c
        double precision, dimension(*), intent(inout) :: work
        double precision, dimension(2,0:dk1-1) :: b2
        integer :: ik1,i2,i3

        do i3=0,pn3-1
            do i2=0,n2-1
                call dzfft(-1, n1, 1., a(0,i2,i3), b2, work(lbnd1:ubnd1), work(lbnd5:ubnd5), 0)

                ! we transpose the subarray that has just been FFTed while the data are still in the cache
                do ik1=0,nk1-1
                    c(:,i2,ik1,i3) = b2(:,ik1)
                end do
            end do
        end do

    end subroutine


! *******************************************
    subroutine fwd_fft_step2(a,c,work)
! *******************************************

        implicit none

        double precision, dimension(2,0:d2-1,0:dk1-1,0:pd3-1), intent(in) :: a
        double precision, dimension(2,0:pd3-1,0:dk1-1,0:pdk2-1,0:np-1), intent(out) :: c
        double precision, dimension(*), intent(inout) :: work
        double precision, dimension(2,0:dk2-1) :: b2
        integer :: i3,ik1,ik2,ip

        if (dealiasing) then
            do ik1=0,nk1-1
                if (trunc_k1(ik1)) then
                    do i3=0,pn3-1
                      call zzfft(-1, n2, 1., a(1,0,ik1,i3), b2, work(lbnd2:ubnd2), work(lbnd5:ubnd5), 0)

                        do ik2=0,pnk2-1
                            do ip=0,np-1
                                c(:,i3,ik1,ik2,ip) = b2(:,ip+np*ik2)
                            end do
                        end do
                    end do
                end if
            end do
        else
            do ik1=0,nk1-1
                do i3=0,pn3-1
                    call zzfft(-1, n2, 1., a(1,0,ik1,i3), b2, work(lbnd2:ubnd2), work(lbnd5:ubnd5), 0)

                    do ik2=0,pnk2-1
                        do ip=0,np-1
                            c(:,i3,ik1,ik2,ip) = b2(:,ip+np*ik2)
                        end do
                    end do
                end do
            end do
        end if

    end subroutine


! *******************************************
    subroutine fwd_fft_step3(a,c,work)
! *******************************************

        implicit none

        double precision, dimension(2,0:pd3-1,0:dk1-1,0:pdk2-1,0:np-1), intent(in) :: a
        double precision, dimension(2,0:dk3-1,0:dk1-1,0:pdk2-1), intent(out) :: c
        double precision, dimension(*), intent(inout) :: work
        double precision, dimension(2,0:dk3-1) :: b2,b3
        integer :: ik1,ik2,ik3,i3,ip

        if (dealiasing) then
            do ik2=0,pnk2-1
                do ik1=0,nk1-1
                    if (trunc_k1k2(ik1,ik2)) then
                        do ip=0,np-1
                            do i3=0,pn3-1
                                b2(:,i3+ip*pn3) = a(:,i3,ik1,ik2,ip)
                            end do
                        end do

                        call zzfft(-1, n3, 1., b2, b3, work(lbnd3:ubnd3), work(lbnd5:ubnd5), 0)

                        do ik3=0,nk3-1
                            c(:,ik3,ik1,ik2) = scaling*b3(:,ik3)
                        end do
                    end if
                end do
            end do
        else
            do ik2=0,pnk2-1
                do ik1=0,nk1-1

                    do ip=0,np-1
                        do i3=0,pn3-1
                            b2(:,i3+ip*pn3) = a(:,i3,ik1,ik2,ip)
                        end do
                    end do

                    call zzfft(-1, n3, 1., b2, b3, work(lbnd3:ubnd3), work(lbnd5:ubnd5), 0)

                    do ik3=0,nk3-1
                        c(:,ik3,ik1,ik2) = scaling*b3(:,ik3)
                    end do
                end do
            end do
        end if
    end subroutine


! *******************************************
    subroutine bck_fft_step1(a,c,work)
! *******************************************

        implicit none

        double precision, dimension(2,0:dk3-1,0:dk1-1,0:pdk2-1), intent(in) :: a
        double precision, dimension(2,0:pdk2-1,0:dk1-1,0:pd3-1,0:np-1), intent(out) :: c
        double precision, dimension(*), intent(inout) :: work
        double precision, dimension(2,0:dk3-1) :: b2
        integer :: i3,ik1,ik2,ip

        if (dealiasing) then

            ! It insures that the modes that are originally truncated in the array a remain truncated after the 1D-FFT
            c = 0.

            do ik1=0, nk1-1
                do ik2=0,pnk2-1
                    if (trunc_k2k1(ik2,ik1)) then
                        call zzfft(1, n3, 1., a(1,0,ik1,ik2), b2, work(lbnd3:ubnd3), work(lbnd5:ubnd5), 0)

                        ! we transpose the subarray that has just been FFTed while the data are still in the cache
                        do ip=0,np-1
                            do i3=0,pn3-1
                                c(:,ik2,ik1,i3,ip) = b2(:,i3+ip*pn3)
                            end do
                        end do
                    end if
                end do
            end do
        else
            do ik1=0, nk1-1
                do ik2=0,pnk2-1
                    call zzfft(1, n3, 1., a(1,0,ik1,ik2), b2, work(lbnd3:ubnd3), work(lbnd5:ubnd5), 0)

                    ! we transpose the subarray that has just been FFTed while the data are still in the cache
                    do ip=0,np-1
                        do i3=0,pn3-1
                            c(:,ik2,ik1,i3,ip) = b2(:,i3+ip*pn3)
                        end do
                    end do
                end do
            end do
        end if

    end subroutine


! *******************************************
    subroutine bck_fft_step2(a,c,work)
! *******************************************

        implicit none

        double precision, dimension(2,0:pdk2-1,0:dk1-1,0:pd3-1,0:np-1), intent(in) :: a
        double precision, dimension(2,0:dk1-1,0:d2-1,0:pd3-1), intent(out) :: c
        double precision, dimension(*), intent(inout) :: work
        double precision, dimension(2,0:dk2-1) :: b2,b3
        integer :: i2,i3,ik1,ik2,ip

        if (dealiasing) then

            ! It insures that the modes that are originally truncated in the array a remain truncated after the 1D-FFT
            c = 0.

            do i3=0,pn3-1
                do ik1=0,nk1-1
                    if (trunc_k1(ik1)) then
                        do ip=0,np-1
                            do ik2=0,pnk2-1
                                b2(:,ip+np*ik2) = a(:,ik2,ik1,i3,ip)
                            end do
                        end do

                        call zzfft(1, n2, 1., b2, b3, work(lbnd2:ubnd2), work(lbnd5:ubnd5), 0)

                        do i2=0,n2-1
                            c(:,ik1,i2,i3) = b3(:,i2)
                        end do
                    end if
                end do
           end do
        else
            do i3=0,pn3-1
                do ik1=0,nk1-1

                    do ip=0,np-1
                        do ik2=0,pnk2-1
                            b2(:,ip+np*ik2) = a(:,ik2,ik1,i3,ip)
                        end do
                    end do

                    call zzfft(1, n2, 1., b2, b3, work(lbnd2:ubnd2), work(lbnd5:ubnd5), 0)

                    do i2=0,n2-1
                        c(:,ik1,i2,i3) = b3(:,i2)
                    end do
                end do
           end do
        end if

    end subroutine


! *******************************************
    subroutine bck_fft_step3(a,work)
! *******************************************

        implicit none

        double precision, dimension(2,0:dk1-1,0:d2-1,0:pd3-1), intent(inout) :: a
        double precision, dimension(*), intent(inout) :: work
        integer :: i2,i3

        do i3=0,pn3-1
            do i2=0,n2-1
                call zdfft(1, n1, 1., a(1,0,i2,i3), a(1,0,i2,i3), work(lbnd4:ubnd4), work(lbnd5:ubnd5), 0)
            end do
        end do

    end subroutine

end module



#elif defined(JMFFT_MPI)
! #######################################################
!
!     Library JMFFT MPI - MPI version for 3D (and 2D) FFT using the JMFFT library real -> complex
!                         Warning: this is a "transposed"  3D FFT ; the 3 dimensions are transposed between the physical and
!                         spectral space. More details are given in the headers of bck_fft and fwd_fft subroutines.
! #######################################################

module fft

    implicit none

    private

    character(len=70), save, public :: fft_lib_name
    public :: init_fft, fwd_fft, bck_fft, init_dealiasing

    ! Logical dimensions of the 3D FFT
    integer, save :: n1,n2,n3

    ! Physical dimensions of the array
    integer, save :: d1,d2,d3

    ! bounds of the working arrays
    integer, save :: lbnd0,ubnd0,lbnd1,ubnd1,lbnd2,ubnd2,lbnd3,ubnd3,lbnd4,ubnd4,lbnd5,ubnd5

    ! number of MPI processes
    integer, save :: np

    ! scaling factor
    double precision, save :: scaling

    ! number of modes and physical dimensions in the spectral space
    integer, save :: nk1,nk2,nk3,dk1,dk2,dk3

    ! In MPI parallelism, the data are splitted between the processes. Each processor has locally only a fraction of the array,
    ! of the modes etc. All the quantities below correspond to these "local" numbers.
    integer, save :: pn1, pn2, pn3
    integer, save :: pnk1, pnk2, pnk3
    integer, save :: pd1, pd2, pd3
    integer, save :: pdk1, pdk2, pdk3

    ! when de-aliasing is used, all the FFT do not have to be calculated. These array store the modes to calculate.
    logical, save :: dealiasing
    logical, dimension(:,:), allocatable :: trunc_k1k2, trunc_k2k1
    logical, dimension(:), allocatable :: trunc_k1

    include "mpif.h"


    contains

! We include all the JMFFT library files
#include "JMFFT-8.0/src/jmccfft.f90"
#include "JMFFT-8.0/src/jmccfft2d.f90"
#include "JMFFT-8.0/src/jmccfft3d.f90"
#include "JMFFT-8.0/src/jmccfftm.f90"
#include "JMFFT-8.0/src/jmccm1d.f90"
#include "JMFFT-8.0/src/jmccm1d2.f90"
#include "JMFFT-8.0/src/jmccm1d3.f90"
#include "JMFFT-8.0/src/jmccm1d4.f90"
#include "JMFFT-8.0/src/jmccm1d5.f90"
#include "JMFFT-8.0/src/jmccm1dp.f90"
#include "JMFFT-8.0/src/jmcctranspcs.f90"
#include "JMFFT-8.0/src/jmcfftmlt.f90"
#include "JMFFT-8.0/src/jmcftfax.f90"
#include "JMFFT-8.0/src/jmcsfft.f90"
#include "JMFFT-8.0/src/jmcsfft2d.f90"
#include "JMFFT-8.0/src/jmcsfft3d.f90"
#include "JMFFT-8.0/src/jmcsfftm.f90"
#include "JMFFT-8.0/src/jmcsm1d.f90"
#include "JMFFT-8.0/src/jmcsm1dxy.f90"
#include "JMFFT-8.0/src/jmdecoup.f90"
#include "JMFFT-8.0/src/jmdecoup3.f90"
#include "JMFFT-8.0/src/jmerreur1.f90"
#include "JMFFT-8.0/src/jmerreur2.f90"
#include "JMFFT-8.0/src/jmerreur3.f90"
#include "JMFFT-8.0/src/jmerreur4.f90"
#include "JMFFT-8.0/src/jmfact.f90"
#include "JMFFT-8.0/src/jmfftfax.f90"
#include "JMFFT-8.0/src/jmgetcode.f90"
#include "JMFFT-8.0/src/jmgeterreur.f90"
#include "JMFFT-8.0/src/jmgeterror.f90"
#include "JMFFT-8.0/src/jmgetmessage.f90"
#include "JMFFT-8.0/src/jmgetnwork.f90"
#include "JMFFT-8.0/src/jmgetsetcode.f90"
#include "JMFFT-8.0/src/jmgetseterreur.f90"
#include "JMFFT-8.0/src/jmgetseterror.f90"
#include "JMFFT-8.0/src/jmgetsetnwork.f90"
#include "JMFFT-8.0/src/jmrfftmlt.f90"
#include "JMFFT-8.0/src/jmscfft.f90"
#include "JMFFT-8.0/src/jmscfft2d.f90"
#include "JMFFT-8.0/src/jmscfft3d.f90"
#include "JMFFT-8.0/src/jmscfftm.f90"
#include "JMFFT-8.0/src/jmscm1d.f90"
#include "JMFFT-8.0/src/jmscm1dxy.f90"
#include "JMFFT-8.0/src/jmsetcode.f90"
#include "JMFFT-8.0/src/jmseterreur.f90"
#include "JMFFT-8.0/src/jmseterror.f90"
#include "JMFFT-8.0/src/jmsetnwork.f90"
#include "JMFFT-8.0/src/jmtable.f90"
#include "JMFFT-8.0/src/jmtransp.f90"
#include "JMFFT-8.0/src/jmsinfftmlt.f90"
#include "JMFFT-8.0/src/jmcosfftmlt.f90"
#include "JMFFT-8.0/src/jmsinfftfax.f90"
#include "JMFFT-8.0/src/jmcosfftfax.f90"
#include "JMFFT-8.0/src/jmtablesin.f90"
#include "JMFFT-8.0/src/jmtablesincos.f90"

! *******************************************
    subroutine init_fft(b,an1,an2,an3,ad1,ad2,ad3,anp,work)
! *******************************************
    !****f* fft/init_fft
    !  NAME
    !    init_fft
    !
    !  DESCRIPTION
    !    Initialize the FFT library
    !
    !  INPUTS
    !    b: field whose dimensions are those of the FFT that are to be performed later.
    !    an1,an2,an3: logical dimensions of the FFT along the first, second and third directions.
    !                 The FFT real -> complex which is the 1st of the 3 required 1D FFT is performed along the 1st direction.
    !                 Because of conjugate relations, only half of the modes along this 1st direction are stored in spectral space.
    !
    !    ad1,ad2,ad3: physical dimensions of the array. d1 >= 2*(n1/2+1), d2 >= n2, d3 >= n3
    !    work: working array whose size must be >= 2*d1*d2*d3/np + 4*8*(n1+n2+n3)+4*100 + 4*max(4+4*n1,8*n2,8*n3)
    !    np: number of MPI processes
    !***

        implicit none

        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(*), intent(inout) :: work
        integer, intent(in) :: an1,an2,an3,ad1,ad2,ad3,anp

        fft_lib_name = 'JMFFT-8.0 - MPI 3D transposed FFT'

        n1=an1
        n2=an2
        n3=an3
        d1=ad1
        d2=ad2
        d3=ad3
        np=anp

        scaling = 1./(n1*n2*n3)

        dealiasing = .false.

        nk1=n1/2+1
        nk2=n2
        nk3=n3
        dk1=d1/2
        dk2=d2
        dk3=d3

        pn1=n1/np
        pn2=n2/np
        pn3=n3/np
        pnk1=nk1/np
        pnk2=nk2/np
        pnk3=nk3/np

        pd1=d1/np
        pd2=d2/np
        pd3=d3/np
        pdk1=dk1/np
        pdk2=dk2/np
        pdk3=dk3/np

        ! the library JMFFT requires two working arrays while only one is given in argument of the subroutine init_fft.
        ! We create artificially two arrays into our single array "work" by determining lower and upper bounds of inner arrays.
        lbnd0 = 1
        ubnd0 = 2*d1*d2*pd3

        lbnd1 = ubnd0 + 1
        ubnd1 = ubnd0 + 100+4*n1

        lbnd2 = ubnd1 + 1
        ubnd2 = ubnd1 + 100+8*n2

        lbnd3 = ubnd2 + 1
        ubnd3 = ubnd2 + 100+8*n3

        lbnd4 = ubnd3 + 1
        ubnd4 = ubnd3 + 100+4*n1

        lbnd5 = ubnd4 + 1
        ubnd5 = ubnd4 + max(4+4*n1,8*n2,8*n3)

        call scfft(0, n1, 1., b, b, work(lbnd1:ubnd1), work(lbnd5:ubnd5), 0)
        call ccfft(0, n2, 1., b, b, work(lbnd2:ubnd2), work(lbnd5:ubnd5), 0)
        call ccfft(0, n3, 1., b, b, work(lbnd3:ubnd3), work(lbnd5:ubnd5), 0)
        call csfft(0, n1, 1., b, b, work(lbnd4:ubnd4), work(lbnd5:ubnd5), 0)

    end subroutine

! *******************************************
    subroutine init_dealiasing(active, trunc_k3k1k2)
! *******************************************
    !****f* fft/init_dealiasing
    !  NAME
    !    init_dealiasing
    !
    !  DESCRIPTION
    !    Activate or deactivate the calculation of modes that are not useful because of dealiasing. It can save time
    !    if a large number of modes are truncated.
    !
    !  INPUTS
    !    active: .true. (dealiasing on) or .false. (dealiasing off)
    !    trunc_k3k1k2: array that contains the modes that are kept in spectral space (.true.) or truncated (.false.)
    !                  the order is the "transposed" order. So if the field is (x1,x2,x3/np) in physical space, the
    !                  array trunc_k3k1k2 is given in order (xk3,xk1,xk2/np)
    !***

        implicit none

        logical, intent(in) :: active
        logical, dimension(0:dk3-1,0:dk1-1,0:pdk2-1), intent(in), optional :: trunc_k3k1k2
        logical, dimension(0:dk1-1) :: tmp_trunc_k1
        integer :: code

        if (allocated(trunc_k1k2)) then
            deallocate(trunc_k1k2)
        end if

        if (allocated(trunc_k2k1)) then
            deallocate(trunc_k2k1)
        end if

        if (allocated(trunc_k1)) then
            deallocate(trunc_k1)
        end if


        if (active .and. present(trunc_k3k1k2)) then
            dealiasing = .true.

            allocate(trunc_k1k2(0:dk1-1,0:pdk2-1))
            allocate(trunc_k2k1(0:pdk2-1,0:dk1-1))
            allocate(trunc_k1(0:dk1-1))

            ! we determine which modes should be calculated in the plane (k1,pk2).
            ! If any mode (k1,pk2) along direction k3 is .true. then the corresponding k3-1D-FFT should be calculated.
            ! The transposition of trunc_k1k2 is stored only to allow a better cache memory access in bck_fft_step1
            trunc_k1k2 = any(trunc_k3k1k2,1)
            trunc_k2k1 = transpose(trunc_k1k2)

            ! we determine which modes should be calculated on the axis k1.
            ! If any mode k1 along direction k2 is .true. in the array trunc_k1k2 then the corresponding k2-1D FFT should be done.
            ! We start with a local reduction (logical .and.) followed by a global reduction between processes.
            tmp_trunc_k1 = any(trunc_k1k2,2)
            call Mpi_allreduce(tmp_trunc_k1,trunc_k1,dk1,MPI_LOGICAL,MPI_LOR ,MPI_COMM_WORLD ,code)

        else
            dealiasing = .false.
        end if

    end subroutine


! *******************************************
    subroutine fwd_fft(b,work)
! *******************************************
    !****f* fft/fwd_fft
    !  NAME
    !    fwd_fft
    !
    !  DESCRIPTION
    !    Perform a forward FFT
    !
    !  INPUTS
    !    b(x1,x2,x3/np): field in the physical space. The third dimension is splitted between the different processes.
    !    work: working array
    !
    !  OUTPUT
    !    b(2,xk3,xk1,xk2/np): field in the spectral space.
    !                         The real and imaginary part are stored contiguously.
    !                         This is a "transposed" FFT: dimensions 1,2 and 3 are not in the same order in the spectral and
    !                         physical space. The last dimension is splitted between the different processes.
    !***

        implicit none
        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(0:d1-1,0:d2-1,0:pd3-1,*), intent(inout) :: work
        integer :: code

        call fwd_fft_step1(b,work(:,:,:,1),work)

        call fwd_fft_step2(work(:,:,:,1),work(:,:,:,2),work)

        call MPI_alltoall(work(:,:,:,2),d1*d2*d3/np**2,MPI_DOUBLE_PRECISION,work(:,:,:,1),d1*d2*d3/np**2, &
                          MPI_DOUBLE_PRECISION,MPI_COMM_WORLD,code)

        call fwd_fft_step3(work(:,:,:,1),b,work)

    end subroutine


! *******************************************
    subroutine bck_fft(b,work)
! *******************************************
    !****f* fft/bck_fft
    !  NAME
    !    bck_fft
    !
    !  DESCRIPTION
    !    Perform a backward FFT. This is the exact inverse of fwd_fft i.e. ffw_fft(bck_fft(a)) = a
    !
    !  INPUTS
    !    b(2,xk3,xk1,xk2/np): field in the spectral space.
    !                         The real and imaginary part are stored contiguously.
    !                         The third dimension is splitted between the different processes.
    !
    !    work: working array
    !
    !  OUTPUT
    !    b(x1,x2,x3/np): field in the physical space.
    !                    The third dimension is splitted between the different processes.
    !***

        implicit none

        double precision, dimension(*), intent(inout) :: b
        double precision, dimension(0:d1-1,0:d2-1,0:pd3-1,*), intent(inout) :: work
        integer :: code

        call bck_fft_step1(b,work(:,:,:,1),work)

        call MPI_alltoall(work(:,:,:,1),d1*d2*d3/np**2,MPI_DOUBLE_PRECISION,work(:,:,:,2),d1*d2*d3/np**2, &
                          MPI_DOUBLE_PRECISION,MPI_COMM_WORLD,code)

        call bck_fft_step2(work(:,:,:,2),b,work)

        call bck_fft_step3(b,work)

    end subroutine

! *******************************************
    subroutine fwd_fft_step1(a,c,work)
! *******************************************

        implicit none

        double precision, dimension(0:d1-1,0:d2-1,0:pd3-1), intent(in) :: a
        double precision, dimension(2,0:d2-1,0:dk1-1,0:pd3-1), intent(out) :: c
        double precision, dimension(*), intent(inout) :: work
        double precision, dimension(2,0:dk1-1) :: b2
        integer :: ik1,i2,i3

        do i3=0,pn3-1
            do i2=0,n2-1
                call scfft(-1, n1, 1., a(0,i2,i3), b2, work(lbnd1:ubnd1), work(lbnd5:ubnd5), 0)

                ! we transpose the subarray that has just been FFTed while the data are still in the cache
                do ik1=0,nk1-1
                    c(:,i2,ik1,i3) = b2(:,ik1)
                end do
            end do
        end do

    end subroutine


! *******************************************
    subroutine fwd_fft_step2(a,c,work)
! *******************************************

        implicit none

        double precision, dimension(2,0:d2-1,0:dk1-1,0:pd3-1), intent(in) :: a
        double precision, dimension(2,0:pd3-1,0:dk1-1,0:pdk2-1,0:np-1), intent(out) :: c
        double precision, dimension(*), intent(inout) :: work
        double precision, dimension(2,0:dk2-1) :: b2
        integer :: i3,ik1,ik2,ip

        if (dealiasing) then
            do ik1=0,nk1-1
                if (trunc_k1(ik1)) then
                    do i3=0,pn3-1
                        call ccfft(-1, n2, 1., a(1,0,ik1,i3), b2, work(lbnd2:ubnd2), work(lbnd5:ubnd5), 0)

                        do ik2=0,pnk2-1
                            do ip=0,np-1
                                c(:,i3,ik1,ik2,ip) = b2(:,ip+np*ik2)
                            end do
                        end do
                    end do
                end if
            end do
        else
            do ik1=0,nk1-1
                do i3=0,pn3-1
                    call ccfft(-1, n2, 1., a(1,0,ik1,i3), b2, work(lbnd2:ubnd2), work(lbnd5:ubnd5), 0)

                    do ik2=0,pnk2-1
                        do ip=0,np-1
                            c(:,i3,ik1,ik2,ip) = b2(:,ip+np*ik2)
                        end do
                    end do
                end do
            end do
        end if

    end subroutine


! *******************************************
    subroutine fwd_fft_step3(a,c,work)
! *******************************************

        implicit none

        double precision, dimension(2,0:pd3-1,0:dk1-1,0:pdk2-1,0:np-1), intent(in) :: a
        double precision, dimension(2,0:dk3-1,0:dk1-1,0:pdk2-1), intent(out) :: c
        double precision, dimension(*), intent(inout) :: work
        double precision, dimension(2,0:dk3-1) :: b2,b3
        integer :: ik1,ik2,ik3,i3,ip

        if (dealiasing) then
            do ik2=0,pnk2-1
                do ik1=0,nk1-1
                    if (trunc_k1k2(ik1,ik2)) then
                        do ip=0,np-1
                            do i3=0,pn3-1
                                b2(:,i3+ip*pn3) = a(:,i3,ik1,ik2,ip)
                            end do
                        end do

                        call ccfft(-1, n3, 1., b2, b3, work(lbnd3:ubnd3), work(lbnd5:ubnd5), 0)

                        do ik3=0,nk3-1
                            c(:,ik3,ik1,ik2) = scaling*b3(:,ik3)
                        end do
                    end if
                end do
            end do
        else
            do ik2=0,pnk2-1
                do ik1=0,nk1-1

                    do ip=0,np-1
                        do i3=0,pn3-1
                            b2(:,i3+ip*pn3) = a(:,i3,ik1,ik2,ip)
                        end do
                    end do

                    call ccfft(-1, n3, 1., b2, b3, work(lbnd3:ubnd3), work(lbnd5:ubnd5), 0)

                    do ik3=0,nk3-1
                        c(:,ik3,ik1,ik2) = scaling*b3(:,ik3)
                    end do
                end do
            end do
        end if
    end subroutine


! *******************************************
    subroutine bck_fft_step1(a,c,work)
! *******************************************

        implicit none

        double precision, dimension(2,0:dk3-1,0:dk1-1,0:pdk2-1), intent(in) :: a
        double precision, dimension(2,0:pdk2-1,0:dk1-1,0:pd3-1,0:np-1), intent(out) :: c
        double precision, dimension(*), intent(inout) :: work
        double precision, dimension(2,0:dk3-1) :: b2
        integer :: i3,ik1,ik2,ip

        if (dealiasing) then

            ! It insures that the modes that are originally truncated in the array a remain truncated after the 1D-FFT
            c = 0.

            do ik1=0, nk1-1
                do ik2=0,pnk2-1
                    if (trunc_k2k1(ik2,ik1)) then
                        call ccfft(1, n3, 1., a(1,0,ik1,ik2), b2, work(lbnd3:ubnd3), work(lbnd5:ubnd5), 0)

                        ! we transpose the subarray that has just been FFTed while the data are still in the cache
                        do ip=0,np-1
                            do i3=0,pn3-1
                                c(:,ik2,ik1,i3,ip) = b2(:,i3+ip*pn3)
                            end do
                        end do
                    end if
                end do
            end do
        else
            do ik1=0, nk1-1
                do ik2=0,pnk2-1
                    call ccfft(1, n3, 1., a(1,0,ik1,ik2), b2, work(lbnd3:ubnd3), work(lbnd5:ubnd5), 0)

                    ! we transpose the subarray that has just been FFTed while the data are still in the cache
                    do ip=0,np-1
                        do i3=0,pn3-1
                            c(:,ik2,ik1,i3,ip) = b2(:,i3+ip*pn3)
                        end do
                    end do
                end do
            end do
        end if

    end subroutine


! *******************************************
    subroutine bck_fft_step2(a,c,work)
! *******************************************

        implicit none

        double precision, dimension(2,0:pdk2-1,0:dk1-1,0:pd3-1,0:np-1), intent(in) :: a
        double precision, dimension(2,0:dk1-1,0:d2-1,0:pd3-1), intent(out) :: c
        double precision, dimension(*), intent(inout) :: work
        double precision, dimension(2,0:dk2-1) :: b2,b3
        integer :: i2,i3,ik1,ik2,ip

        if (dealiasing) then

            ! It insures that the modes that are originally truncated in the array a remain truncated after the 1D-FFT
            c = 0.

            do i3=0,pn3-1
                do ik1=0,nk1-1
                    if (trunc_k1(ik1)) then
                        do ip=0,np-1
                            do ik2=0,pnk2-1
                                b2(:,ip+np*ik2) = a(:,ik2,ik1,i3,ip)
                            end do
                        end do

                        call ccfft(1, n2, 1., b2, b3, work(lbnd2:ubnd2), work(lbnd5:ubnd5), 0)

                        do i2=0,n2-1
                            c(:,ik1,i2,i3) = b3(:,i2)
                        end do
                    end if
                end do
           end do
        else
            do i3=0,pn3-1
                do ik1=0,nk1-1

                    do ip=0,np-1
                        do ik2=0,pnk2-1
                            b2(:,ip+np*ik2) = a(:,ik2,ik1,i3,ip)
                        end do
                    end do

                    call ccfft(1, n2, 1., b2, b3, work(lbnd2:ubnd2), work(lbnd5:ubnd5), 0)

                    do i2=0,n2-1
                        c(:,ik1,i2,i3) = b3(:,i2)
                    end do
                end do
           end do
        end if

    end subroutine


! *******************************************
    subroutine bck_fft_step3(a,work)
! *******************************************

        implicit none

        double precision, dimension(2,0:dk1-1,0:d2-1,0:pd3-1), intent(inout) :: a
        double precision, dimension(*), intent(inout) :: work
        integer :: i2,i3

        do i3=0,pn3-1
            do i2=0,n2-1
                call csfft(1, n1, 1., a(1,0,i2,i3), a(1,0,i2,i3), work(lbnd4:ubnd4), work(lbnd5:ubnd5), 0)
            end do
        end do

    end subroutine

end module


#endif
