!******************************************************************************
!
! NS3D: a Navier-Stokes solver for incompressible and stratified flow
!
! Organization : LadHyX, Ecole Polytechnique, France
!
! Permission is granted to copy and distribute this file or modified
! versions of this file for no fee, provided the copyright notice and
! this permission notice are preserved on all copies.
!
!******************************************************************************

#include "config.h"
#include "extended_config.h"

!****************************************************************************
program NS3Dstrat
!***************************************************************************
!****c* ns3d/main
!  NAME
!    NS3Dstrat
!
!  DESCRIPTION
!    This is the main program that solves the incompressible Navier-Stockes equations with a
!    pseudo-spectral method.
!***

    ! External modules used by the program
    use timing
    use fft
    use global_vars
    use subfunctions
    use time_scheme
    use input
    use output
    use gen_velocity

#ifdef MPI
    use mpi_times
#endif

    implicit none

#ifdef MPI
    call mpi_init(code)
    call mpi_comm_rank(MPI_COMM_WORLD,rank,code)
    call mpi_time(0)
#endif

    ! we time the whole program
    call timer_start("main")

    call write_text
    call write_text('###################################################')
    call write_text('            PROGRAM NS3D  '//program_version)
    call write_text('###################################################')
    call write_text


    ! ******************************
    !  INITIALIZATION
    ! ******************************

    ! we time the initilization process but without timing more precisely the nested subroutines
    call timer_start("initialization")
    call timer_set_active(.false.)

    call write_text('----------------------------------------------------------------------------------','(A79)')
    call write_text('INITIALIZATION')
    call write_text('----------------------------------------------------------------------------------','(A79)')

    ! we set the timestep counter to 0
    it = 0

    ! we read the global variables from the file "data.in"
    call read_param

#ifdef MPI
    if (rank==0) then
        write(*,*) '----------------------------'
        write(*,'(A29,TR3,I5,TR1,A1,TR1,I5,TR1,A1,TR1,I5)') ' DIMENSIONS OF THE SIMULATION:', gnx,'x',gny,'x',gnz
        write(*,'(A29,TR3,I5)') ' NUMBER OF MPI PROCESSES', nb_proc
        write(*,*) '----------------------------'

        ! we display the global variables
        call print_param
    end if
#else
    write(*,*) '----------------------------'
    write(*,'(A29,TR3,I5,TR1,A1,TR1,I5,TR1,A1,TR1,I5)') ' DIMENSIONS OF THE SIMULATION:', gnx,'x',gny,'x',gnz
    write(*,*) '----------------------------'

    ! we display the global variables
    call print_param
#endif

    ! coordinates of the colocation points
    call init_coord(xx,yy,zz)

    ! wavenumbers
    call init_k(kx,kx2,ky,ky2,kz,kz2,kxmax,kxmax2,kymax,kymax2,kzmax,kzmax2)

    ! initialization of the de_aliasing table
    if (de_aliasing) then
        call init_trunctab(trunctab)
    end if

    ! initialization of the Fast Fourier Transform (FFT) functions
    call write_text
    call write_text('Initialization of the Fast Fourier Transformation (FFT)')
#ifdef MPI
    call init_fft(vox,gnx,gny,gnz,gdx,gdy,gdz,nb_proc,fft_work)
    
    ! when de-aliasing is used, all the spectral modes do not have to be calculated since they will be set to 0 anyway.    
    ! We can skip the calculation of these modes by with "init_dealiasing(.true., trunctab)" where trunctab stores
    ! the modes to calculate.
    ! In this case additionnal logical tests are performed. De-aliasing saves time when the amount of modes not to be 
    ! computed is important and when dealing with large dimensions. Otherwise, it can be slower because of the
    ! additionnal logical tests. In most situations, init_dealiasing(.false.,trunctab) is faster.
    call init_dealiasing(.false., trunctab)

#else
    call init_fft(vox,gnx,gny,gnz,gdx,gdy,gdz,fft_work)
#endif
    call write_text('  -> '//fft_lib_name)


    ! we generate the base state and we write it into a file
    call write_text
    call write_text('Base state')
    call write_text("  -> "//base2D_type)
    call gen_base2D(vbx,vby,vbz,wbx,wby,wbz,base2D_type,work)
    call output_base(vbx,vby,vbz,wbx,wby,wbz)

    ! generation of the initial velocity
    call write_text
    call write_text('Initial velocity')
    call write_text("  -> "//velo_type)
    call gen_velo(vox,voy,voz,rho,velo_type,work)
				
    if (white_noise /= 0.) then
        call write_text('  -> white noise')
        call add_white_noise(vox,voy,voz,white_noise)
    end if

    ! we switch into spectral space
    call fwd_fft(vox,fft_work)
    call fwd_fft(voy,fft_work)
    call fwd_fft(voz,fft_work)
    if (stratification) then
        call fwd_fft(rho,fft_work)
    end if

    ! Projection to impose incompressibility and de-aliasing by de_aliasing   
    call projection(vox,voy,voz)
    if(de_aliasing) then       
        call trunc(vox,voy,voz,rho)       
    endif

    ! we save the initial fields
    call output2(vox,voy,voz,rho,work(:,1),work(:,2),work(:,3),work(:,4))
    call write_text
    
    call timer_set_active(.true.)
    call timer_stop("initialization")

    ! ******************************
    ! TIME STEPPING
    ! ******************************
    
    call write_text
    call write_text('----------------------------------------------------------------------------------','(A79)')
    call write_text('TIME STEPPING: '//time_scheme_name)
    call write_text('----------------------------------------------------------------------------------','(A79)')
    call write_text

    ! we enter the main loop with the values of fields at instant it-1 and we exit the loop with the values at instant it
    do it=1,itmax
        call timer_start("time_stepping")
        
        ! time stepping scheme
        call timer_start("time_scheme")
        call time_step(vox,voy,voz,rho,time_scheme_name,work)
        call timer_stop("time_scheme")

        ! Projection to impose incompressibility and de-aliasing by de_aliasing
        call timer_start("projection")
        call projection(vox,voy,voz)
        call timer_stop("projection")

        if(de_aliasing) then
            call timer_start("de_aliasing")
            call trunc(vox,voy,voz,rho)
            call timer_stop("de_aliasing")
        endif

        ! we call the output subfunctions at specified instants
        ! the max(*,1) in mod() avoids division by 0 (some compilers test the second condition even if the first one is .false.)
        if (output1_period /= 0 .and. mod(it, max(output1_period,1)) == 0) then
            call timer_start("output1")
            call output1(vox,voy,voz,rho)
            call timer_stop("output1")
        end if

        if (output2_period /= 0 .and. mod(it, max(output2_period,1)) == 0) then
            call timer_start("output2")
            call output2(vox,voy,voz,rho,work(:,1),work(:,2),work(:,3),work(:,4))
            call timer_stop("output2")
        end if

        if (output3_period /= 0 .and. mod(it, max(output3_period,1)) == 0) then
            call timer_start("output3")
            call output3(vox,voy,voz,rho,work(:,1),work(:,2),work(:,3),work(:,4),work(:,5),work(:,6),work(:,7),work(:,8))
            call timer_stop("output3")
        end if

        ! We check if "Not a Numbers" (NaN) are present in the results. If so, we stop the simulation
        ! Warning: the function "isnan" is not part of the Fortran 90/95 norm and is not implemented in all compilers
#ifdef TEST_NAN
        call timer_start("test_nan")
        if (detect_nan(vox)) then
            call write_text('ERROR : a Not a Number is present in the results. Failure of the simulation.')
            call timer_stop("test_nan")
            exit
        end if
        call timer_stop("test_nan")
#endif
        
        call timer_stop("time_stepping")
    end do
    call timer_set_active(.false.)
    
    call write_text
    call write_text('----------------------------------------------------------------------------------','(A79)')
    call write_text('END OF THE SIMULATION')
    call write_text('----------------------------------------------------------------------------------','(A79)')
    call write_text

    ! we save all the final fields (instant itmax)
    it = itmax
    call output2(vox,voy,voz,rho,work(:,1),work(:,2),work(:,3),work(:,4))
    
    call timer_set_active(.true.)
    call timer_stop("main")

    call write_text
    call timer_report_all

#ifdef MPI
    call mpi_time(1)
    call mpi_finalize(code)
#endif

end program
