!******************************************************************************
!
! NS3D: a Navier-Stokes solver for incompressible and stratified flow
!
! Organik3ation : LadHyX, Ecole Polytechnique, France
!
! Permission is granted to copy and distribute this file or modified
! versions of this file for no fee, provided the copyright notice and
! this permission notice are preserved on all copies.
!
!******************************************************************************

#include "config.h"
#include "extended_config.h"

!***************************************************************************
module time_scheme
!***************************************************************************
!****h* ns3d/time_scheme
!  NAME
!    time_scheme
!
!  DESCRIPTION
!    Implement different time advancement schemes: Runge and Kutta of
!    order 2, 3 and 4 as well as Adams-Bashforth schemes of order 2 and 3.
!    These schemes are applied on velocity and density fields in spectral space.
!***

    use global_vars
    use subfunctions

    implicit none

    logical, save :: initialized = .false.

    contains

    subroutine time_step(svx,svy,svz,srho,scheme_type,work)
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(inout) :: svx, svy, svz, srho
        character(len=*), intent(in) :: scheme_type
        double precision, dimension(dx*dy*dz,*), intent(inout) :: work

        ! Adams-Bashforth of order 2
        if (scheme_type=='Adams-Bashforth of order 2' .and. initialized) then
                call ab2(svx,svy,svz,srho,work(:,1),work(:,2),work(:,3),work(:,4),work(:,5),work(:,6),work(:,7),work(:,8))

        else if (scheme_type=='Adams-Bashforth of order 2' .and. (.not. initialized)) then
                call init_ab2(svx,svy,svz,srho,work(:,1),work(:,2),work(:,3),work(:,4),work(:,5),work(:,6),work(:,7),work(:,8))
                call precal_exp_fact(dt,eex,eexs,eey,eeys,eez,eezs)
                initialized = .true.


        ! Runge and Kutta of order 2
        else if (scheme_type=='Runge and Kutta of order 2' .and. initialized) then
                call rk2(svx,svy,svz,srho,work(:,1),work(:,2),work(:,3),work(:,4),work(:,5),work(:,6),work(:,7),work(:,8))

        else if (scheme_type=='Runge and Kutta of order 2' .and. (.not. initialized)) then
                call precal_exp_fact(dt/2.,eex,eexs,eey,eeys,eez,eezs)
                call rk2(svx,svy,svz,srho,work(:,1),work(:,2),work(:,3),work(:,4),work(:,5),work(:,6),work(:,7),work(:,8))
                initialized = .true.


        ! Runge and Kutta of order 3
        else if (scheme_type=='Runge and Kutta of order 3' .and. initialized) then
            call rk3(svx,svy,svz,srho,work(:,1),work(:,2),work(:,3),work(:,4),work(:,5),work(:,6),work(:,7),work(:,8), &
                     work(:,9),work(:,10),work(:,11),work(:,12))
        else if (scheme_type=='Runge and Kutta of order 3' .and. (.not. initialized)) then
            call precal_exp_fact(dt/3.,eex,eexs,eey,eeys,eez,eezs)
            call rk3(svx,svy,svz,srho,work(:,1),work(:,2),work(:,3),work(:,4),work(:,5),work(:,6),work(:,7),work(:,8), &
                     work(:,9),work(:,10),work(:,11),work(:,12))
            initialized = .true.


        ! Runge and Kutta of order 4
        else if (scheme_type=='Runge and Kutta of order 4' .and. initialized) then
            call rk4(svx,svy,svz,srho,work(:,1),work(:,2),work(:,3),work(:,4),work(:,5),work(:,6),work(:,7),work(:,8), &
                         work(:,9),work(:,10),work(:,11),work(:,12))
        else if (scheme_type=='Runge and Kutta of order 4' .and. (.not. initialized)) then
            call precal_exp_fact(dt/2.,eex,eexs,eey,eeys,eez,eezs)
            call rk4(svx,svy,svz,srho,work(:,1),work(:,2),work(:,3),work(:,4),work(:,5),work(:,6),work(:,7),work(:,8), &
                     work(:,9),work(:,10),work(:,11),work(:,12))
            initialized = .true.

        end if

    end subroutine


! *************************************************************************
    subroutine rk2(svx,svy,svz,srho,wkx1,wky1,wkz1,wkrho1,wkx2,wky2,wkz2,wkrho2)
! *************************************************************************
    !****f* time_scheme/rk2
    !  NAME
    !    rk2
    !
    !  DESCRIPTION
    !    perform a time step with a Runge and Kutta scheme of order 2
    !
    !  INPUTS
    !    svx,svy,svz,srho: velocity and density fields at time t in spectral space
    !
    !  OUTPUT
    !    svx,svy,svz,srho: velocity and density fields at time t+dt in spectral space
    !    wkx1,wky1,wkz1,wkrho1,wkx2,wky2,wkz2,wkrho2: the working fields are modified
    !***

        implicit none

        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(inout) :: svx, svy, svz, srho
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(out) :: wkx1,wky1,wkz1,wkrho1
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(out) :: wkx2,wky2,wkz2,wkrho2
        double precision :: ee,ees,ee2,ees2, dt2
        integer :: ik1, ik2, ik3

        wkx1=svx
        wky1=svy
        wkz1=svz
        wkrho1 = srho

        call timer_start("nonlin_term")
        call non_linear_term(wkx1,wky1,wkz1,wkrho1,wkx2,wky2,wkz2,wkrho2)
        call timer_stop("nonlin_term")

        dt2 = dt/2.

        do ik3=0,nk3-1
            do ik2=0,nk2-1
                do ik1=0,nk1-1
                    ee=eex(IKX)*eey(IKY)*eez(IKZ)
                    ees=eexs(IKX)*eeys(IKY)*eezs(IKZ)

                    wkx1(:,ik1,ik2,ik3)=(svx(:,ik1,ik2,ik3) + dt2*wkx2(:,ik1,ik2,ik3))*ee
                    wky1(:,ik1,ik2,ik3)=(svy(:,ik1,ik2,ik3) + dt2*wky2(:,ik1,ik2,ik3))*ee
                    wkz1(:,ik1,ik2,ik3)=(svz(:,ik1,ik2,ik3) + dt2*wkz2(:,ik1,ik2,ik3))*ee
                    wkrho1(:,ik1,ik2,ik3)=(srho(:,ik1,ik2,ik3) + dt2*wkrho2(:,ik1,ik2,ik3))*ees
                end do
            end do
        end do

        call timer_start("nonlin_term")
        call non_linear_term(wkx1,wky1,wkz1,wkrho1,wkx2,wky2,wkz2,wkrho2)
        call timer_stop("nonlin_term")

        do ik3=0,nk3-1
            do ik2=0,nk2-1
                do ik1=0,nk1-1
                    ee = eex(IKX)*eey(IKY)*eez(IKZ)
                    ee2 = ee*ee
                    ees = eexs(IKX)*eeys(IKY)*eezs(IKZ)
                    ees2 = ees*ees

                    svx(:,ik1,ik2,ik3)=svx(:,ik1,ik2,ik3)*ee2 + dt*wkx2(:,ik1,ik2,ik3)*ee
                    svy(:,ik1,ik2,ik3)=svy(:,ik1,ik2,ik3)*ee2 + dt*wky2(:,ik1,ik2,ik3)*ee
                    svz(:,ik1,ik2,ik3)=svz(:,ik1,ik2,ik3)*ee2 + dt*wkz2(:,ik1,ik2,ik3)*ee
                    srho(:,ik1,ik2,ik3)=srho(:,ik1,ik2,ik3)*ees2 + dt*wkrho2(:,ik1,ik2,ik3)*ees
                end do
            end do
        end do

    end subroutine


! *************************************************************************
    subroutine rk3(svx,svy,svz,srho,wkx1,wky1,wkz1,wkrho1,wkx2,wky2,wkz2,wkrho2,wkx3,wky3,wkz3,wkrho3)
! *************************************************************************
    !****f* time_scheme/rk3
    !  NAME
    !    rk3
    !
    !  DESCRIPTION
    !    perform a time step with a Runge and Kutta scheme of order 3
    !
    !  INPUTS
    !    svx,svy,svz,srho: velocity and density fields at time t in spectral space
    !
    !  OUTPUT
    !    svx,svy,svz,srho: velocity and density fields at time t+dt in spectral space
    !    wkx1,wky1,wkz1,wkrho1,wkx2,wky2,wkz2,wkrho2,wkx3,wky3,wkz3,wkrho3: the working fields are modified
    !***

        implicit none

        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(inout) :: svx, svy, svz, srho
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(out) :: wkx1,wky1,wkz1,wkrho1
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(out) :: wkx2,wky2,wkz2,wkrho2
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(out) :: wkx3,wky3,wkz3,wkrho3
        double precision :: ee,ees,ee2,ees2,ee3,ees3, dt3, dt4, dt23, dt34
        integer :: ik1, ik2, ik3

        wkx1=svx
        wky1=svy
        wkz1=svz
        wkrho1 = srho

        call timer_start("nonlin_term")
        call non_linear_term(wkx1,wky1,wkz1,wkrho1,wkx2,wky2,wkz2,wkrho2)
        call timer_stop("nonlin_term")

        dt3 = dt/3.

        do ik3=0,nk3-1
            do ik2=0,nk2-1
                do ik1=0,nk1-1
                    ee=eex(IKX)*eey(IKY)*eez(IKZ)
                    ees=eexs(IKX)*eeys(IKY)*eezs(IKZ)

                    wkx1(:,ik1,ik2,ik3)=(svx(:,ik1,ik2,ik3) + dt3*wkx2(:,ik1,ik2,ik3))*ee
                    wky1(:,ik1,ik2,ik3)=(svy(:,ik1,ik2,ik3) + dt3*wky2(:,ik1,ik2,ik3))*ee
                    wkz1(:,ik1,ik2,ik3)=(svz(:,ik1,ik2,ik3) + dt3*wkz2(:,ik1,ik2,ik3))*ee
                    wkrho1(:,ik1,ik2,ik3)=(srho(:,ik1,ik2,ik3) + dt3*wkrho2(:,ik1,ik2,ik3))*ees
                end do
            end do
        end do

        call timer_start("nonlin_term")
        call non_linear_term(wkx1,wky1,wkz1,wkrho1,wkx3,wky3,wkz3,wkrho3)
        call timer_stop("nonlin_term")

        dt23 = 2.*dt/3.
        dt4 = dt/4.

        do ik3=0,nk3-1
            do ik2=0,nk2-1
                do ik1=0,nk1-1
                    ee=eex(IKX)*eey(IKY)*eez(IKZ)
                    ees=eexs(IKX)*eeys(IKY)*eezs(IKZ)
                    ee2 = ee*ee
                    ees2 = ees*ees
                    ee3 = ee2*ee
                    ees3 = ees2*ees

                    wkx1(:,ik1,ik2,ik3)=svx(:,ik1,ik2,ik3)*ee2 + dt23*wkx3(:,ik1,ik2,ik3)*ee
                    wky1(:,ik1,ik2,ik3)=svy(:,ik1,ik2,ik3)*ee2 + dt23*wky3(:,ik1,ik2,ik3)*ee
                    wkz1(:,ik1,ik2,ik3)=svz(:,ik1,ik2,ik3)*ee2 + dt23*wkz3(:,ik1,ik2,ik3)*ee
                    wkrho1(:,ik1,ik2,ik3)=srho(:,ik1,ik2,ik3)*ees2 + dt23*wkrho3(:,ik1,ik2,ik3)*ees

                    svx(:,ik1,ik2,ik3)=(svx(:,ik1,ik2,ik3) + dt4*wkx2(:,ik1,ik2,ik3))*ee3
                    svy(:,ik1,ik2,ik3)=(svy(:,ik1,ik2,ik3) + dt4*wky2(:,ik1,ik2,ik3))*ee3
                    svz(:,ik1,ik2,ik3)=(svz(:,ik1,ik2,ik3) + dt4*wkz2(:,ik1,ik2,ik3))*ee3
                    srho(:,ik1,ik2,ik3)=(srho(:,ik1,ik2,ik3) + dt4*wkrho2(:,ik1,ik2,ik3))*ees3
                end do
            end do
        end do

        call timer_start("nonlin_term")
        call non_linear_term(wkx1,wky1,wkz1,wkrho1,wkx3,wky3,wkz3,wkrho3)
        call timer_stop("nonlin_term")

        dt34 = 3.*dt/4.

        do ik3=0,nk3-1
            do ik2=0,nk2-1
                do ik1=0,nk1-1
                    ee=eex(IKX)*eey(IKY)*eez(IKZ)
                    ees=eexs(IKX)*eeys(IKY)*eezs(IKZ)

                    svx(:,ik1,ik2,ik3)=svx(:,ik1,ik2,ik3) + dt34*wkx3(:,ik1,ik2,ik3)*ee
                    svy(:,ik1,ik2,ik3)=svy(:,ik1,ik2,ik3) + dt34*wky3(:,ik1,ik2,ik3)*ee
                    svz(:,ik1,ik2,ik3)=svz(:,ik1,ik2,ik3) + dt34*wkz3(:,ik1,ik2,ik3)*ee
                    srho(:,ik1,ik2,ik3)=srho(:,ik1,ik2,ik3) + dt34*wkrho3(:,ik1,ik2,ik3)*ees
                end do
            end do
        end do

    end subroutine

! *************************************************************************
    subroutine rk4(svx,svy,svz,srho,wkx1,wky1,wkz1,wkrho1,wkx2,wky2,wkz2,wkrho2,wkx3,wky3,wkz3,wkrho3)
! *************************************************************************
    !****f* time_scheme/rk4
    !  NAME
    !    rk4
    !
    !  DESCRIPTION
    !    perform a time step with a Runge and Kutta scheme of order 4
    !
    !  INPUTS
    !    svx,svy,svz,srho: velocity and density fields at time t in spectral space
    !
    !  OUTPUT
    !    svx,svy,svz,srho: velocity and density fields at time t+dt in spectral space
    !    wkx1,wky1,wkz1,wkrho1,wkx2,wky2,wkz2,wkrho2,wkx3,wky3,wkz3,wkrho3: the working fields are modified
    !***

        implicit none

        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(inout) :: svx, svy, svz, srho
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(out) :: wkx1,wky1,wkz1,wkrho1
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(out) :: wkx2,wky2,wkz2,wkrho2
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(out) :: wkx3,wky3,wkz3,wkrho3
        double precision :: ee,ees,ee2,ees2, dt2, dt3, dt6
        integer :: ik1, ik2, ik3

        dt2 = dt/2.
        dt3 = dt/3.
        dt6 = dt/6.

        wkx1=svx
        wky1=svy
        wkz1=svz
        wkrho1 = srho

        call timer_start("nonlin_term")
        call non_linear_term(wkx1,wky1,wkz1,wkrho1,wkx2,wky2,wkz2,wkrho2)
        call timer_stop("nonlin_term")

        do ik3=0,nk3-1
            do ik2=0,nk2-1
                do ik1=0,nk1-1
                    ee=eex(IKX)*eey(IKY)*eez(IKZ)
                    ees=eexs(IKX)*eeys(IKY)*eezs(IKZ)

                    wkx1(:,ik1,ik2,ik3)=(svx(:,ik1,ik2,ik3) + dt2*wkx2(:,ik1,ik2,ik3))*ee
                    wky1(:,ik1,ik2,ik3)=(svy(:,ik1,ik2,ik3) + dt2*wky2(:,ik1,ik2,ik3))*ee
                    wkz1(:,ik1,ik2,ik3)=(svz(:,ik1,ik2,ik3) + dt2*wkz2(:,ik1,ik2,ik3))*ee
                    wkrho1(:,ik1,ik2,ik3)=(srho(:,ik1,ik2,ik3) + dt2*wkrho2(:,ik1,ik2,ik3))*ees
                end do
            end do
        end do
        
        call timer_start("nonlin_term")
        call non_linear_term(wkx1,wky1,wkz1,wkrho1,wkx3,wky3,wkz3,wkrho3)
        call timer_stop("nonlin_term")

        do ik3=0,nk3-1
            do ik2=0,nk2-1
                do ik1=0,nk1-1
                    ee=eex(IKX)*eey(IKY)*eez(IKZ)
                    ees=eexs(IKX)*eeys(IKY)*eezs(IKZ)
                    ee2 = ee*ee
                    ees2 = ees*ees

                    wkx1(:,ik1,ik2,ik3)=svx(:,ik1,ik2,ik3)*ee + dt2*wkx3(:,ik1,ik2,ik3)
                    wky1(:,ik1,ik2,ik3)=svy(:,ik1,ik2,ik3)*ee + dt2*wky3(:,ik1,ik2,ik3)
                    wkz1(:,ik1,ik2,ik3)=svz(:,ik1,ik2,ik3)*ee + dt2*wkz3(:,ik1,ik2,ik3)
                    wkrho1(:,ik1,ik2,ik3)=srho(:,ik1,ik2,ik3)*ees + dt2*wkrho3(:,ik1,ik2,ik3)

                    wkx2(:,ik1,ik2,ik3)=(svx(:,ik1,ik2,ik3)+dt6*wkx2(:,ik1,ik2,ik3))*ee2 + dt3*wkx3(:,ik1,ik2,ik3)*ee
                    wky2(:,ik1,ik2,ik3)=(svy(:,ik1,ik2,ik3)+dt6*wky2(:,ik1,ik2,ik3))*ee2 + dt3*wky3(:,ik1,ik2,ik3)*ee
                    wkz2(:,ik1,ik2,ik3)=(svz(:,ik1,ik2,ik3)+dt6*wkz2(:,ik1,ik2,ik3))*ee2 + dt3*wkz3(:,ik1,ik2,ik3)*ee
                    wkrho2(:,ik1,ik2,ik3)=(srho(:,ik1,ik2,ik3)+dt6*wkrho2(:,ik1,ik2,ik3))*ees2 + dt3*wkrho3(:,ik1,ik2,ik3)*ees
                end do
            end do
        end do

        call timer_start("nonlin_term")
        call non_linear_term(wkx1,wky1,wkz1,wkrho1,wkx3,wky3,wkz3,wkrho3)
        call timer_stop("nonlin_term")
        
        do ik3=0,nk3-1
            do ik2=0,nk2-1
                do ik1=0,nk1-1
                    ee=eex(IKX)*eey(IKY)*eez(IKZ)
                    ees=eexs(IKX)*eeys(IKY)*eezs(IKZ)
                    ee2 = ee*ee
                    ees2 = ees*ees

                    wkx1(:,ik1,ik2,ik3)=svx(:,ik1,ik2,ik3)*ee2 + dt*wkx3(:,ik1,ik2,ik3)*ee
                    wky1(:,ik1,ik2,ik3)=svy(:,ik1,ik2,ik3)*ee2 + dt*wky3(:,ik1,ik2,ik3)*ee
                    wkz1(:,ik1,ik2,ik3)=svz(:,ik1,ik2,ik3)*ee2 + dt*wkz3(:,ik1,ik2,ik3)*ee
                    wkrho1(:,ik1,ik2,ik3)=srho(:,ik1,ik2,ik3)*ees2 + dt*wkrho3(:,ik1,ik2,ik3)*ees

                    wkx2(:,ik1,ik2,ik3)=wkx2(:,ik1,ik2,ik3) +  dt3*wkx3(:,ik1,ik2,ik3)*ee
                    wky2(:,ik1,ik2,ik3)=wky2(:,ik1,ik2,ik3) +  dt3*wky3(:,ik1,ik2,ik3)*ee
                    wkz2(:,ik1,ik2,ik3)=wkz2(:,ik1,ik2,ik3) +  dt3*wkz3(:,ik1,ik2,ik3)*ee
                    wkrho2(:,ik1,ik2,ik3)=wkrho2(:,ik1,ik2,ik3) + dt3*wkrho3(:,ik1,ik2,ik3)*ees
                end do
            end do
        end do

        call timer_start("nonlin_term")
        call non_linear_term(wkx1,wky1,wkz1,wkrho1,wkx3,wky3,wkz3,wkrho3)
        call timer_stop("nonlin_term")

        do ik3=0,nk3-1
            do ik2=0,nk2-1
                do ik1=0,nk1-1
                    svx(:,ik1,ik2,ik3)=wkx2(:,ik1,ik2,ik3) + dt6*wkx3(:,ik1,ik2,ik3)
                    svy(:,ik1,ik2,ik3)=wky2(:,ik1,ik2,ik3) + dt6*wky3(:,ik1,ik2,ik3)
                    svz(:,ik1,ik2,ik3)=wkz2(:,ik1,ik2,ik3) + dt6*wkz3(:,ik1,ik2,ik3)
                    srho(:,ik1,ik2,ik3)=wkrho2(:,ik1,ik2,ik3) + dt6*wkrho3(:,ik1,ik2,ik3)
               end do
            end do
        end do
        
    end subroutine


! *************************************************************************
    subroutine ab2(svx,svy,svz,srho,wkx1,wky1,wkz1,wkrho1,wkx2,wky2,wkz2,wkrho2)
! *************************************************************************
    !****f* time_scheme/ab2
    !  NAME
    !    ab2
    !
    !  DESCRIPTION
    !    perform a time step with an Adams-Bashforth scheme of order 2
    !
    !  INPUTS
    !    svx,svy,svz,srho: velocity and density fields at time t in spectral space
    !
    !  OUTPUT
    !    svx,svy,svz,srho: velocity and density fields at time t+dt in spectral space
    !
    !  NOTES
    !    When calling this subroutine to evaluate vo(t+1), the global working fields must be initialized this way:
    !      * vo : vo(t) in spectral space
    !      * wk1 : /
    !      * wk2 : non_linear_term{vo(t-1)} in spectral space
    !***

        implicit none

        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(inout) :: svx, svy, svz, srho
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(out) :: wkx1,wky1,wkz1,wkrho1
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(inout) :: wkx2,wky2,wkz2,wkrho2
        double precision :: ee,ees,ee2,ees2, dt2, dt32
        integer :: ik1, ik2, ik3

        dt2 = dt/2.

        do ik3=0,nk3-1
            do ik2=0,nk2-1
                do ik1=0,nk1-1
                    ee=eex(IKX)*eey(IKY)*eez(IKZ)
                    ees=eexs(IKX)*eeys(IKY)*eezs(IKZ)
                    ee2 = ee*ee
                    ees2 = ees*ees

                    wkx1(:,ik1,ik2,ik3)=svx(:,ik1,ik2,ik3)*ee - dt2*wkx2(:,ik1,ik2,ik3)*ee2
                    wky1(:,ik1,ik2,ik3)=svy(:,ik1,ik2,ik3)*ee - dt2*wky2(:,ik1,ik2,ik3)*ee2
                    wkz1(:,ik1,ik2,ik3)=svz(:,ik1,ik2,ik3)*ee - dt2*wkz2(:,ik1,ik2,ik3)*ee2
                    wkrho1(:,ik1,ik2,ik3)=srho(:,ik1,ik2,ik3)*ees - dt2*wkrho2(:,ik1,ik2,ik3)*ees2
                end do
            end do
        end do

        call timer_start("nonlin_term")
        call non_linear_term(svx,svy,svz,srho,wkx2,wky2,wkz2,wkrho2)
        call timer_stop("nonlin_term")

        dt32 = 3.*dt/2.

        do ik3=0,nk3-1
            do ik2=0,nk2-1
                do ik1=0,nk1-1
                    ee=eex(IKX)*eey(IKY)*eez(IKZ)
                    ees=eexs(IKX)*eeys(IKY)*eezs(IKZ)

                    svx(:,ik1,ik2,ik3)=wkx1(:,ik1,ik2,ik3) + dt32*wkx2(:,ik1,ik2,ik3)*ee
                    svy(:,ik1,ik2,ik3)=wky1(:,ik1,ik2,ik3) + dt32*wky2(:,ik1,ik2,ik3)*ee
                    svz(:,ik1,ik2,ik3)=wkz1(:,ik1,ik2,ik3) + dt32*wkz2(:,ik1,ik2,ik3)*ee
                    srho(:,ik1,ik2,ik3)=wkrho1(:,ik1,ik2,ik3) + dt32*wkrho2(:,ik1,ik2,ik3)*ees
                end do
            end do
        end do

    end subroutine


! *************************************************************************
    subroutine init_ab2(svx,svy,svz,srho,wkx1,wky1,wkz1,wkrho1,wkx2,wky2,wkz2,wkrho2)
! *************************************************************************
    !****f* time_scheme/init_ab2
    !  NAME
    !    init_ab2
    !
    !  DESCRIPTION
    !    Initialization for the Adams-Bashforth scheme of order 2
    !
    !  INPUTS
    !    svx,svy,svz,srho: velocity and density fields at time t=0 in spectral space
    !
    !  OUTPUT
    !    svx,svy,svz,srho: velocity and density field at time t=dt in spectral space
    !***

        implicit none

        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(inout) :: svx, svy, svz, srho
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(out) :: wkx1,wky1,wkz1,wkrho1
        double precision, dimension(2,0:dk1-1,0:dk2-1,0:dk3-1), intent(out) :: wkx2,wky2,wkz2,wkrho2
        double precision :: ee,ees,ee2,ees2, dt2
        integer :: ik1, ik2, ik3

        ! we use a Runge and Kutta scheme of order 2 to perform the first time step and initialize the working fields
        call precal_exp_fact(dt/2.,eex,eexs,eey,eeys,eez,eezs)

        wkx1=svx
        wky1=svy
        wkz1=svz
        wkrho1 = srho

        call timer_start("nonlin_term")
        call non_linear_term(wkx1,wky1,wkz1,wkrho1,wkx2,wky2,wkz2,wkrho2)
        call timer_stop("nonlin_term")

        dt2 = dt/2.

        do ik3=0,nk3-1
            do ik2=0,nk2-1
                do ik1=0,nk1-1
                    ee=eex(IKX)*eey(IKY)*eez(IKZ)
                    ees=eexs(IKX)*eeys(IKY)*eezs(IKZ)

                    wkx1(:,ik1,ik2,ik3)=(svx(:,ik1,ik2,ik3) + dt2*wkx2(:,ik1,ik2,ik3))*ee
                    wky1(:,ik1,ik2,ik3)=(svy(:,ik1,ik2,ik3) + dt2*wky2(:,ik1,ik2,ik3))*ee
                    wkz1(:,ik1,ik2,ik3)=(svz(:,ik1,ik2,ik3) + dt2*wkz2(:,ik1,ik2,ik3))*ee
                    wkrho1(:,ik1,ik2,ik3)=(srho(:,ik1,ik2,ik3) + dt2*wkrho2(:,ik1,ik2,ik3))*ees
                end do
            end do
        end do

        call timer_start("nonlin_term")
        call non_linear_term(wkx1,wky1,wkz1,wkrho1,wkx2,wky2,wkz2,wkrho2)
        call timer_stop("nonlin_term")

        wkx1=svx
        wky1=svy
        wkz1=svz
        wkrho1 = srho

        do ik3=0,nk3-1
            do ik2=0,nk2-1
                do ik1=0,nk1-1
                    ee = eex(IKX)*eey(IKY)*eez(IKZ)
                    ee2 = ee*ee
                    ees = eexs(IKX)*eeys(IKY)*eezs(IKZ)
                    ees2 = ees*ees

                    svx(:,ik1,ik2,ik3)=svx(:,ik1,ik2,ik3)*ee2 + dt*wkx2(:,ik1,ik2,ik3)*ee
                    svy(:,ik1,ik2,ik3)=svy(:,ik1,ik2,ik3)*ee2 + dt*wky2(:,ik1,ik2,ik3)*ee
                    svz(:,ik1,ik2,ik3)=svz(:,ik1,ik2,ik3)*ee2 + dt*wkz2(:,ik1,ik2,ik3)*ee
                    srho(:,ik1,ik2,ik3)=srho(:,ik1,ik2,ik3)*ees2 + dt*wkrho2(:,ik1,ik2,ik3)*ees
                end do
            end do
        end do

        ! The non-linear term at initial time is necessary
        call timer_start("nonlin_term")
        call non_linear_term(wkx1,wky1,wkz1,wkrho1,wkx2,wky2,wkz2,wkrho2)
        call timer_stop("nonlin_term")

    end subroutine

!***************************************************************************
    subroutine precal_exp_fact(dtl,eex,eexs,eey,eeys,eez,eezs)
!***************************************************************************
    !****f* time_scheme/precal_exp_fact
    !  NAME
    !    precal_exp_fact
    !
    !  DESCRIPTION
    !    Calculate the pre-exponential terms required for time integration
    !
    !  INPUTS
    !    dtl: time step of the time integration
    !
    !  OUTPUT
    !    eex,eexs,eey,eeys,eez,eezs: the pre-exponential factors that are global variables used
    !                                for the time advancement
    !***

        implicit none

        double precision, intent(in) :: dtl
        double precision, dimension(0:dkx-1), intent(out) :: eex, eexs
        double precision, dimension(0:dky-1), intent(out) :: eey, eeys
        double precision, dimension(0:dkz-1), intent(out) :: eez, eezs

        eex=exp(-nu*kx2*dtl)
        eexs=exp(-nu/schm*kx2*dtl)

        eey=exp(-nu*ky2*dtl)
        eeys=exp(-nu/schm*ky2*dtl)

        eez=exp(-nu*kz2*dtl)
        eezs=exp(-nu/schm*kz2*dtl)

    end subroutine

end module
